#!/usr/bin/env python

import numpy as np
import os
from scipy import sparse
from time import time

from tqdm import tqdm, trange

from astropy.io import fits

import logging
# Define logging
_log = logging.getLogger(__name__)
_log.setLevel(logging.INFO)

__all__ = ['NMF', 'data_masked_only', 'data_masked_only_revert', 
           'NMFcomponents', 'NMFmodelling', 'NMFsubtraction', 'NMFbff', 'nmf_math']

""" 
nmf.py

    This piece of software is developed and maintained by Guangtun Ben Zhu, 
    It is designed to solve nonnegative matrix factorization (NMF) given a dataset with heteroscedastic 
    uncertainties and missing data with a vectorized multiplicative update rule (Zhu 2016).
    The un-vectorized (i.e., indexed) update rule for NMF without uncertainties or missing data was
    originally developed by Lee & Seung (2000), and the un-vectorized update rule for NMF
    with uncertainties or missing data was originally developed by Blanton & Roweis (2007).

    As all the codes, this code can always be improved and any feedback will be greatly appreciated.

    Note:
      -- Between W and H, which one is the basis set and which one is the coefficient
         depends on how you interpret the data, because you can simply transpose everything
         as in X-WH versus X^T - (H^T)(W^T)
      -- Everything needs to be non-negative

    Here are some small tips for using this code:
      -- The algorithm can handle heteroscedastic uncertainties and missing data.
         You can supply the weight (V) and the mask (M) at the instantiation:

         >> g = nmf.NMF(X, V=V, M=M)

         This can also be very useful if you would like to iterate the process
         so that you can exclude certain new data by updating the mask.
         For example, if you want to perform a 3-sigma clipping after an iteration
         (assuming V is the inverse variance below):

         >> chi2_red, time_used = g.SolveNMF()
         >> New_M = np.copy(M)
         >> New_M[np.fabs(np.sqrt(V)*(X-np.dot(g.W, g.H)))>3] = False
         >> New_g = nmf.NMF(X, V=V, M=New_M)

         Caveat: Currently you need to re-instantiate the object whenever you update
         the weight (V), the mask (M), W, H or n_components.
         At the instantiation, the code makes a copy of everything.
         For big jobs with many iterations, this could be a severe bottleneck.
         For now, I think this is a safer way.

      -- It has W_only and H_only options. If you know H or W, and would like
         to calculate W or H. You can run, e.g.,

         >> chi2_red, time_used = g.SolveNMF(W_only=True)

         to get the other matrix (H in this case).


    Copyright (c) 2015-2016 Guangtun Ben Zhu

    Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
    and associated documentation files (the "Software"), to deal in the Software without 
    restriction, including without limitation the rights to use, copy, modify, merge, publish, 
    distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the 
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all copies or 
    substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
    BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
    DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
class NMF:
    """
    Nonnegative Matrix Factorization - Build a set of nonnegative basis components given 
    a dataset with Heteroscedastic uncertainties and missing data with a vectorized update rule.

    Algorithm:
      -- Iterative multiplicative update rule

    Input: 
      -- X: m x n matrix, the dataset

    Optional Input/Output: 
      -- n_components: desired size of the basis set, default 5

      -- V: m x n matrix, the weight, (usually) the inverse variance
      -- M: m x n binary matrix, the mask, False means missing/undesired data
      -- H: n_components x n matrix, the H matrix, usually interpreted as the coefficients
      -- W: m x n_components matrix, the W matrix, usually interpreted as the basis set
    
    (See README for how to retrieve the test data)
    
    Construct a new basis set with 12 components
    Instantiation: 
        >> g = nmf.NMF(flux, V=ivar, n_components=12)
    Run the solver: 
        >> chi2_red, time_used = g.SolveNMF()

    If you have a basis set, say W, you would like to calculate the coefficients:
        >> g = nmf.NMF(flux, V=ivar, W=W_known)
    Run the solver: 
        >> chi2_red, time_used = g.SolveNMF(H_only=True)

    Comments:
      -- Between W and H, which one is the basis set and which one is the coefficient 
         depends on how you interpret the data, because you can simply transpose everything
         as in X-WH versus X^T - (H^T)(W^T)
      -- Everything needs to be non-negative

    References:
      -- Guangtun Ben Zhu, 2016
         A Vectorized Algorithm for Nonnegative Matrix Factorization with 
         Heteroskedastic Uncertainties and Missing Data
         AJ/PASP, (to be submitted)
      -- Blanton, M. and Roweis, S. 2007
         K-corrections and Filter Transformations in the Ultraviolet, Optical, and Near-infrared
         The Astronomical Journal, 133, 734
      -- Lee, D. D., & Seung, H. S., 2001
         Algorithms for non-negative matrix factorization
         Advances in neural information processing systems, pp. 556-562

    To_do:
      -- 

    History:
        -- 22-May-2016, Documented, BGT, JHU
        -- 13-May-2016, Add projection mode (W_only, H_only), BGT, JHU
        -- 30-Nov-2014, Started, BGT, JHU
    """

    def __init__(self, X, W=None, H=None, V=None, M=None, n_components=5, rand_seed=None):
        """ Initialization

        Parameters
        ----------
        X : ndarray
            m x n matrix, the input data set.

        Keyword Arguments
        -----------------
        W : ndarray, optional
            m x n_components matrix, the W matrix, usually interpreted as the basis set
        H : ndarray, optional
            n_components x n matrix, the H matrix, usually interpreted as the coefficients
        V : ndarray, optional
            m x n matrix, the weight, (usually) the inverse variance
        M : ndarray, optional
            m x n binary matrix, the mask, False means missing/undesired data
        n_components : int, optional
            Desired size of the basis set. Default 5.
            Cannot be larger than the number of columns in X (n).
        rand_seed : int, optional
            random seed for reproducibility of any randomly generated matrices
        """

        # Create local random number generator to avoid global seed setting
        self._rng = np.random.default_rng(seed=rand_seed)

        self._n_components = n_components
        self.maxiters = 1000
        self.tol = 1e-5

        # Make a copy for the safety of everything; should not be a bottleneck
        self.X = np.copy(X) 
        if (np.count_nonzero(self.X<0)>0):
            _log.info("There are negative values in X. Setting them to be zero...")
            self.X[self.X<0] = 0.

        # Basis set
        if (W is None):
            self.W = self._rng.random((self.X.shape[0], n_components))
        else:
            # if (W.shape != (self.X.shape[0], n_components)):
            if W.shape[0] != self.X.shape[0]:
                raise ValueError(f"Initial W has wrong shape {W.shape}.")
            self.W = np.copy(W)

        # Coefficients
        if (H is None):
            self.H = self._rng.random((n_components, self.X.shape[1]))
        else:
            # if (H.shape != (n_components, self.X.shape[1])):
            if H.shape[1] != self.X.shape[1]:
                raise ValueError(f"Initial H has wrong shape {H.shape}")
            self.H = np.copy(H)

        # Weights
        if (V is None):
            self._V = np.ones(self.X.shape)
        else:
            if (V.shape != self.X.shape):
                raise ValueError(f"Initial V(Weight) has wrong shape {V.shape}.")
            self.V = np.copy(V)

        # Mask
        if (M is None):
            self.M = np.ones(self.X.shape, dtype='bool')
        else:
            if (M.shape != self.X.shape):
                raise ValueError(f"M(ask) has wrong shape {M.shape}.")
            if (M.dtype != bool):
                raise TypeError("M(ask) needs to be boolean.")
            self.M = np.copy(M)

        # Set masked elements to be zero
        self._V[~self.M] = 0

    @property
    def n_components(self):
        """ Number of components """
        return self._n_components
    @n_components.setter
    def n_components(self, ncomps_new):
        """ Set the number of components """

        # Add additional components to W and H if needed
        ncomps_full = self._W.shape[1]
        if ncomps_new > ncomps_full:
            # Make new arrays with the new (larger) number of components
            Wtemp = self._rng.random((self.X.shape[0], ncomps_new))
            Htemp = self._rng.random((ncomps_new, self.X.shape[1]))
            # Copy over the old values
            Wtemp[:, :(ncomps_full-1)] = self._W
            Htemp[:(ncomps_full-1), :] = self._H
            # Set the new arrays
            self.W = Wtemp
            self.H = Htemp        

        # Update the number of components
        self._n_components = ncomps_new

    @property
    def W(self):
        """ Basis set components """
        return self._W[:, 0:self.n_components]
    @W.setter
    def W(self, value):
        """ Set the basis set components """
        neg_mask = value<0
        if (np.count_nonzero(neg_mask) > 0):
            _log.warning("There are negative values in W. Setting them to be zero...")
            value[neg_mask] = 0.
        # Fortran ordering, column elements contiguous in memory.
        self._W = np.array(value, order='F') 

    @property
    def H(self):
        """ Coefficients """
        return self._H[0:self.n_components, :]
    @H.setter
    def H(self, value):
        """ Set the coefficients """
        neg_mask = value<0
        if (np.count_nonzero(neg_mask) > 0):
            _log.warning("There are negative values in H. Setting them to be zero...")
            value[neg_mask] = 0.
        # C ordering, row elements contiguous in memory.
        self._H = np.array(value, order = 'C') 

    @property
    def V(self):
        """ Weights """
        return self._V
    @V.setter
    def V(self, value):
        neg_mask = value<0
        if (np.count_nonzero(neg_mask<0) > 0):
            _log.warning("There are negative values in V. Setting them to be zero...")
            value[neg_mask<0] = 0.
        self._V = value

    @property 
    def V_size(self):
        """ Total number of valid data points """
        return np.count_nonzero(self.V)

    @property
    def cost(self):
        """  Total cost of a given set s """
        diff = self.X - np.dot(self.W, self.H)
        # chi2 = np.einsum('ij,ij', self.V*diff, diff) / self.V_size

        # Reduced chi2 metric: chi2 = sum(V*(X-WH)^2) / V_size
        chi2 = np.sum(self.V * diff**2) / self.V_size
        return chi2

    def SolveNMF(self, W_only=False, H_only=False, sparsemode=False, 
                 maxiters=None, tol=None, verbose=False, **kwargs):
        """
        Construct the NMF basis set and coefficients

        Parameters
        ----------
        W_only : bool
            If True, only update W, assuming H is known
        H_only : bool
            If True, only update H, assuming W is known
        sparsemode : bool
            If True, use sparse matrix for the calculation. Benchmarks show that
            this is only faster for very sparse data (e.g., >95% zeros).
        maxiters : int
            Maximum number of iterations. If None, use the default value
            at self.maxiters (initially 1000).
        tol : float
            Convergence criterion. If None, use the default value at self.tol
            (initially 1E-5).
        verbose: bool
            If True, print out extra information such as wall clock time and chi2

        Returns
        -------
        chi2 : float
            Reduced final cost
        time_used : float
            Time in sec to run
        """

        t0 = time()

        if (maxiters is None): 
            maxiters = self.maxiters
        if (tol is None):
            tol = self.tol

        chi2 = self.cost
        oldchi2 = np.finfo(chi2.dtype).max # 1e100

        if (W_only and H_only):
            _log.warning("Both W_only and H_only are set to be True. Returning...")
            return (chi2, 0.)

        if (sparsemode == True):
            V = sparse.csr_matrix(self.V)
            VT = sparse.csr_matrix(self.V.T)
            func_mult = sparse.csr_matrix.multiply
            func_dot = sparse.csr_matrix.dot
        else:
            V = self.V #np.copy(self.V)
            VT = V.T
            func_mult = np.multiply
            func_dot = np.dot

        if (not W_only):
            XVT = func_mult(VT, self.X.T)
        if (not H_only):
            XV = func_mult(V, self.X)

        niter = 0
        # while (niter < self.maxiters) and (chi_change > self.tol):
        tqdm_desc = f"{self.n_components}-component NMF"
        tqdm_iter = trange(maxiters, desc=tqdm_desc, leave=False)
        for niter in tqdm_iter:

            # Update H
            if (not W_only):
                H_up = func_dot(XVT, self.W)
                WHVT = func_mult(VT, np.dot(self.W, self.H).T)
                H_down = func_dot(WHVT, self.W)
                H_new = self.H * (H_up.T / H_down.T)
                self._H[0:self.n_components, :] = H_new

            # Update W
            if (not H_only):
                W_up = func_dot(XV, self.H.T)
                WHV = func_mult(V, np.dot(self.W, self.H))
                W_down = func_dot(WHV, self.H.T)
                W_new = self.W * (W_up / W_down)
                self._W[:, 0:self.n_components] = W_new

            # Update chi2
            oldchi2 = chi2
            chi2 = self.cost
            chi_change = (oldchi2 - chi2) / oldchi2

            # Some quick check. May need its error class ...
            if (not np.isfinite(chi2)):
               raise ValueError("NMF construction failed, likely due to missing data")

            # print output updates
            if (np.mod(niter, 20)==0):
                tqdm_iter.set_description(f"{tqdm_desc}: Chi2={chi2:.3f}, Change={chi_change:.1e}")
                # _log.info(f"Current Chi2={chi2:.4f}, Previous Chi2={oldchi2:.4f}, Change={100*chi_change:.4f}% @ niters={niter}")

            if (np.abs(chi_change) < tol):
                break

        if (niter == maxiters):
            _log.warning(f"SolveNMF reached maximum number of iterations ({maxiters})")

        time_total = time() - t0
        if verbose:
            _log.info(f"{self.n_components} took {time_total:.1f} seconds. Final chi2={chi2:.4f} in {niter+1} iterations.")

        return (chi2, time_total)
    
    def SolveNMF_OneByOne(self, W_only=False, H_only=False, sparsemode=False, 
                          maxiters=None, tol=None, verbose=False, ncomp_start=1, **kwargs):
        """Build NMF components one-by-one
        
        Each component is built by adding one more component to the basis set and the coefficients.
        This can be useful for checking the convergence of the NMF algorithm.
        Option to start at a specific component number in case W and H are already partially built.
        """

        if ncomp_start < 1:
            raise ValueError("ncomp_start must be at least 1")
        elif ncomp_start > self.n_components:
            raise ValueError(f"ncomp_start={ncomp_start} must be less than {self.n_components+1}")
        elif ncomp_start == 1:
            _log.info(f"Solving {self.n_components}-component NMF one-by-one")
        else:
            _log.info(f"Solving {self.n_components}-component NMF one-by-one, starting at {ncomp_start} components")

        time_total = 0
        for ncomp in range(ncomp_start, self.n_components+1):
            self.n_components = ncomp
            chi2, partial_time = self.SolveNMF(W_only=W_only, H_only=H_only, sparsemode=sparsemode, 
                                               maxiters=maxiters, tol=tol, verbose=verbose, **kwargs)
            time_total += partial_time

        if verbose:
            _log.info(f"Took {time_total:.1f} seconds to reach current solution.")

        return (chi2, time_total)


# This code is the nmf_imaging.py adjusted for pyKLIP at https://bitbucket.org/pyKLIP/pyklip/src/master/pyklip/nmf_imaging.py
# Another version is kept at https://github.com/seawander/nmf_imaging/blob/master/nmf_imaging_for_pyKLIP.py

def data_masked_only(data, mask):
    """ Return the data where the same regions are ignored in all the data

    Parameters
    ----------
    data: ndarray
        shape (p, N) where p is the number of pixels in each reference and N is the number of references,
    mask: boolean array
        shape (p) where False will be ignored in the output

    Returns
    -------
    data_focused: ndarray
        shape (p_focused, N) where p_focused is the number of True's in mask
    """   
    data_focused = data[mask]
    if len(data.shape) == 1:
        data_focused = data_focused[:, np.newaxis]
    
    return data_focused
    
def data_masked_only_revert(data, mask):
    """ Return the data where the same regions were ignored in all the data

    Parameters
    ----------
    data : ndarray
        shape (p_focused, N) where N is the number of references, p_focused is the number of 1's in mask
    mask : boolean array
        shape (p) where False was previously ignored in the input

    Returns
    -------
    data_focused_revert: ndarray
        shape (p, N) where  p is the number of pixels in each reference. Ignored data is replaced with NaNs.
    """ 
    data_focused_revert = np.zeros((len(mask), data.shape[1])) * np.nan
    data_focused_revert[mask] = data
    
    return data_focused_revert

def NMFcomponents(ref, ref_err=None, ignore_mask=None, n_components=5, 
                  maxiters=1000, oneByOne=True, save=False, save_name=None, save_dir=None, 
                  recalculate=False, rand_seed=None, **kwargs):
    """Returns the NMF components, where the rows contain the information.

    Parameters
    ----------
    ref : 2D array
        Reference images of shape (N,p) where N is the number of references, p is the number of pixels in each reference.
    ref_err : 2D array
         Uncertainty in the reference images of shape (N,p).
    ignore_mask : 1D or 2D array
        Mask pixels in each image that you don't want to use. Can either ben of size (p) or (N,p).
    n_components : int
        Number of components to be used. If None, all the components will be used. If greater than N, N will be used.
    maxiters : int
        Maximum number of iterations for the NMF algorithm.
    oneByOne : bool
        If True, the components will be built one-by-one. If False, all the components will be built at once.
    save : bool
        If True, the components and coefficients will be saved to disk for reuse.
    save_dir : str
        Path to save the NMF components (files named '_comp.fits' and '_coef.fits')
    recalculate : bool
        If True, the components will be recalculated even if they already exist on disk.
    rand_seed : int
        Random seed for reproducibility of any randomly generated matrices

    Returns
    -------
    components : 2D array
        NMf components of shape (n_components, p).
    """
    
    if save_dir is None:
        save_dir = '.'
    elif not os.path.exists(save_dir):
        raise ValueError(f"Path {save_dir} does not exist.")
    
    name_append = '' if save_name is None else f'{save_name}'
    ext_name = 'onebyone' if oneByOne else 'allatonce'
    fpath_comp = os.path.join(save_dir, f'{name_append}_comp_{ext_name}_{n_components}comps.fits')
    fpath_coef = os.path.join(save_dir, f'{name_append}_coef_{ext_name}_{n_components}comps.fits')

    # Search for file with same name but larger number of components
    if (not os.path.exists(fpath_comp)) and (not recalculate):
        for f in os.listdir(save_dir):
            if f.startswith(f'{name_append}_comp_{ext_name}_') and f.endswith('comps.fits'):
                ncomps_file = int(f.split('_')[-1].split('comps')[0])
                if ncomps_file > n_components:
                    _log.info(f'Found existing components with {ncomps_file} components. Using those instead...')
                    fpath_comp = os.path.join(save_dir, f)
                    fpath_coef = os.path.join(save_dir, f.replace(f'{name_append}_comp_', f'{name_append}_coef_'))
                    break

    # Set recalculate keyword to False if the components do not exist on disk
    # Will still calculate components, but provides correct print messages
    if recalculate and (not os.path.exists(fpath_comp)):
        recalculate = False

    if ref_err is None:
        ref_err = np.sqrt(ref)
        
    ncomp, npix = ref.shape
    if (n_components is None) or (n_components > ncomp):
        n_components = ncomp
        
    if ignore_mask is None:
        ignore_mask = np.ones_like(ref, dtype='bool')
    elif ignore_mask.shape == ref.shape[1]:
        # Create a 2D mask of shape (N, p)
        ignore_mask = np.tile(ignore_mask, (ncomp, 1))

    assert ref.shape == ref_err.shape, "Reference and error maps must have the same shape."
    assert ref.shape == ignore_mask.shape, "Reference and ignore mask must have the same shape."
    
    # ignore certain values in component construction
    ignore_mask[ref <= 0] = False # 1. negative values
    ignore_mask[~np.isfinite(ref)] = False # 2. infinite values
    ignore_mask[np.isnan(ref)] = False # 3. nan values
    
    ignore_mask[ref_err <= 0] = False # 1. negative values in input error map
    ignore_mask[~np.isfinite(ref_err)] = False # 2. infinite values in input error map
    ignore_mask[np.isnan(ref_err)] = False # 3. nan values in input error map
    
    # Ignore pixels that are always masked across all references
    mask_mark = np.bitwise_or.reduce(ignore_mask, axis=0)
    
    # Transpose the data to comply with the NMF class
    ref_columnized = data_masked_only(ref.T, mask_mark)
    ref_err_columnized = data_masked_only(ref_err.T, mask_mark)
    mask_columnized = data_masked_only(ignore_mask.T, mask_mark)

    # Update bad pixels
    ref_columnized[~mask_columnized] = 0 # assign 0 to ignored values, should not impact the final result
    ref_err_columnized[~mask_columnized] = np.nanmax(ref_err_columnized) # assign max uncertainty to ignored values, should not impact the final result

    ref_columnized[np.isnan(ref_columnized)] = 0 # assign 0 to NaN values, should not impact the final result
    ref_err_columnized[np.isnan(ref_err_columnized)] = np.nanmax(ref_err_columnized) # assign max uncertainty to NaN values, should not impact the final result

    # weights
    weights = 1.0 / ref_err_columnized**2


    # component calculation
    skip_calc = False
    if os.path.exists(fpath_comp) and (not recalculate):
        _log.info(f'Loading existing components and coefficients from {save_dir}...')
        W_assign = fits.getdata(fpath_comp)
        H_assign = fits.getdata(fpath_coef)
        if W_assign.shape[1] == n_components:
            skip_calc = True
        if W_assign.shape[1] > n_components:
            _log.info(f'You have already had {W_assign.shape[1]} components while asking for {n_components}. Returning...')
            W_assign = W_assign[:, :n_components]
            H_assign = H_assign[:n_components, :]
            skip_calc = True
        else:
            # Set oneByOne to True to get remaining components
            _log.info(f'{W_assign.shape[1]} existing components lower than requested ({n_components}). Calculating the rest...')
            oneByOne = True
    else:
        W_assign = None
        H_assign = None

    if recalculate or (not os.path.exists(fpath_comp)):
        _log.info("Building components one-by-one...")
        if recalculate:
            _log.warning('Recalculating components from scratch...')
        elif not os.path.exists(fpath_comp):
            _log.info(f'{fpath_comp} does not exist, calculating from scratch...')

    if not skip_calc:
        g_img = NMF(ref_columnized, W=W_assign, H=H_assign, V=weights, M=mask_columnized, 
                    n_components=n_components, rand_seed=rand_seed)
        
        # Ensure maxiters is an integer
        g_img.maxiters = int(maxiters)

        if oneByOne:
            ncomp_start = 1 if W_assign is None else W_assign.shape[1]
            chi2, time_used = g_img.SolveNMF_OneByOne(ncomp_start=ncomp_start, **kwargs)
        else:
            chi2, time_used = g_img.SolveNMF(**kwargs)
        W_assign, H_assign = (g_img.W, g_img.H)

        if save:
            str_write = f"overwriting" if recalculate else "writing"
            _log.info(f'\t\t\t Finished NMF for {n_components} components, {str_write} 2D component matrix to {fpath_comp}')
            fits.writeto(fpath_comp, W_assign, overwrite=True)

            _log.info(f'\t\t\t Finished NMF for {n_components} components, {str_write} 2D coefficient matrix to {fpath_coef}')
            fits.writeto(fpath_coef, H_assign, overwrite = True)

    # Normalize components
    norm_val = np.sqrt(np.nansum(W_assign**2, axis=0))
    components_column = W_assign / norm_val
    # components of shape (p, n_components)
    components = data_masked_only_revert(components_column, mask=mask_mark)   

    return components.T
    

def NMFmodelling(sci, components, n_components=None, mask_components=None, mask_data_imputation=None, 
                 sci_err=None, oneByOne=False, maxiters=1e3, sciThresh=0, rand_seed=None, **kwargs):
    """ NMF modeling of target data

    Parameters:
    -----------
    sci: 1D array
        Target image of shape (p) where p is the number of pixels
    components: 2D array
        NMF components of shape (N, p)
    n_components: int
        Number of components to be used. If None, all the components will be used. If greater than N, N will be used.
    mask_components: 1D array
        Mask pixels in component array that you don't want to use. Should be of size (p).
        False values will be ignored in the modeling.
    mask_data_imputation: 1D array
        Mask pixels in the target data that you don't want to use. Can be of size (p).
        False values will be ignored in the modeling.

    Args:
        sci: 1D array, p pixels
        components: N * p, calculated using NMFcomponents.
        n_components: how many components do you want to use. If None, all the components will be used.
        cube: whether output a cube or not (increasing the number of components).
        sciThresh: ignore the regions with low photon counts. Especially when they are ~10^-15 or smaller. I chose 0 in this case.
    
    Returns:
        NMF model of the target.
    """

    if n_components is None:
        n_components = components.shape[0]
    n_components = np.min([n_components, components.shape[0]])
        
    if sci_err is None:
        sci_err = np.sqrt(sci)

    if mask_components is None:
        masks_nans = np.isnan(components[:n_components])
        mask_components = ~np.bitwise_or.reduce(masks_nans, axis=0)

     # columnize the components, make sure NonnegMFPy returns correct results.
    components_column_all = data_masked_only(components[:n_components].T, mask=mask_components)  
    # Ensure components are normalized
    norm_val = np.sqrt(np.nansum(components_column_all**2, axis = 0))
    components_column_all = components_column_all / norm_val
    
    if mask_data_imputation is not None:
        flag_di = True
        _log.info('Data Imputation!')
    else:
        flag_di = False
        mask_data_imputation = np.ones(sci.shape, dtype='bool')
    
    # Ignore pixels that are False in either mask (ie., True in both masks)
    mask = mask_components & mask_data_imputation # will be used for modeling

    # Set values below threshold to 0
    sci[sci < sciThresh] = 0
    sci_err[sci==0] = np.nanmax(sci_err)

    # Mask invalid pixels
    mask[sci <= 0] = False
    mask[np.isnan(sci)] = False
    mask[~np.isfinite(sci)] = False
    
    # Columnize the target and its error and keep only the valid pixels
    sci_column = data_masked_only(sci, mask=mask)
    sci_err_column = data_masked_only(sci_err, mask=mask)
    weights = 1.0 / sci_err_column**2
    components_column = data_masked_only(components[:n_components].T, mask=mask)

    sci_img = NMF(sci_column, V=weights, W=components_column, 
                  n_components=n_components, rand_seed=rand_seed)
    if oneByOne:
        chi2, time_used = sci_img.SolveNMF_OneByOne(H_only=True, maxiters=maxiters, **kwargs)
    else:
        chi2, time_used = sci_img.SolveNMF(H_only=True, maxiters=maxiters, **kwargs)

    coefs = sci_img.H
    if flag_di: # do data imputation
        model_column = np.dot(components_column_all, coefs)
        model = data_masked_only_revert(model_column, mask_components)
        model[~mask_components] = np.nan
    else: # do not do data imputation
        model_column = np.dot(components_column, coefs)
        model = data_masked_only_revert(model_column, mask)
        model[~mask] = np.nan

    return model.flatten() #model_column.T.flatten()
    
def NMFsubtraction(sci, model, frac=1):
    """NMF subtraction with a correction factor, frac."""
    frac = np.atleast_1d(frac)

    result = []
    for i, fraction in enumerate(frac):
        result.append(sci - model*fraction)

    return np.array(result).squeeze()
    
def NMFbff(sci, model, fracs=None, **kwargs):
    """BFF subtraction.
    Args:
        sci:
        model:
        fracs: (if need to be).
    Returns: 
        best frac
    """

    from webbpsf_ext import robust
    
    if fracs is None:
        fracs = np.arange(0.50, 1.001, 0.001)

    # std_infos = []
    # for i, frac in enumerate(fracs):
    #     data_slice = sci - model*frac
    #     while 1:
    #         if np.nansum(data_slice > np.nanmedian(data_slice) + 3*np.nanstd(data_slice)) == 0 or np.nansum(data_slice < np.nanmedian(data_slice) -3*np.nanstd(data_slice)) == 0: # Modified from -10 on 2018/07/12
    #             break
    #         data_slice[data_slice > np.nanmedian(data_slice) + 3*np.nanstd(data_slice)] = np.nan
    #         data_slice[data_slice < np.nanmedian(data_slice) - 3*np.nanstd(data_slice)] = np.nan # Modified from -10 on 2018/07/12
    #     std_info = np.nanstd(data_slice)
    #     std_infos.append(std_info)
    # std_infos = np.array(std_infos)

    std_infos = np.array([robust.medabsdev(sci - model*frac) for frac in fracs])
    ind_min = np.where(std_infos == np.nanmin(std_infos))[0]
    return fracs[ind_min]   
   
def nmf_math(sci, ref_psfs, sci_err=None, ref_psfs_err=None, 
             ignore_mask=None, mask_data_imputation=None, sci_type='disk', 
             n_components=10, maxiters=5000, oneByOne=True, rand_seed=None,
             save=True, save_name=None, save_dir=None, recalculate=False, **kwargs):
    """ Main NMF function for high contrast imaging.

    Parameters
    ----------
    sci : 1D or 2D array
        Target image, dimension: height * width.

    Args:  
        sci (1D array): target image, dimension: height * width.
        refs (2D array): reference cube, dimension: referenceNumber * height * width.
        sci_err, ref_err: uncertainty for sci and refs, repectively. If None is given, the squareroot of the two arrays will be adopted.
    
        componentNum (integer): number of components to be used. Default: 5. Caution: choosing too many components will slow down the computation.
        maxiters (integer): number of iterations needed. Default: 10^5.
        oneByOne (boolean): whether to construct the NMF components one-by-one. Default: True.
        sci_type (string,  default: "disk" or "d" for circumsetllar disks by Bin Ren, the user can use "planet" or "p" for planets): are we aiming at finding circumstellar disks or planets?
    Returns: 
        result (1D array): NMF modeling result. Only the final subtraction result is returned.
    """

    rng = np.random.default_rng(seed=rand_seed)

    sh_sci = sci.shape
    sh_ref = ref_psfs.shape

    if len(sh_sci) == 2:
        sci = sci.ravel()
    if len(sh_ref) == 3:
        ref_psfs = ref_psfs.reshape(sh_ref[0], -1)

    # Total number of ref components and pixels
    ncomp_tot, npix = ref_psfs.shape

    # Ensure n_components is not greater than the number of references
    if n_components > ncomp_tot:
        _log.warning(f"Requested n_components={n_components} is greater than number of references ({ncomp_tot}). Setting n_components={ncomp_tot}.")
        n_components = ncomp_tot

    if sci_err is not None:
        sci_err = sci_err.reshape(sci.shape)
    if ref_psfs_err is not None:
        ref_psfs_err = ref_psfs_err.reshape(ref_psfs.shape)

    if (ignore_mask is not None):
        if ignore_mask.size == npix:
            ignore_mask = ignore_mask.ravel()
        elif ignore_mask.size == ncomp_tot * npix:
            ignore_mask = ignore_mask.reshape(ncomp_tot, npix)
        else:
            raise ValueError("ignore_mask must be same size as `sci` or `ref_psfs`.")
        
    if mask_data_imputation is not None:
        mask_data_imputation = mask_data_imputation.reshape(sci.shape)

    # Generate components
    kwargs_components = {
        'ref_err': ref_psfs_err, 'ignore_mask': ignore_mask, 
        'n_components': n_components, 'maxiters': maxiters, 'oneByOne': oneByOne, 
        'save': save, 'save_name': save_name, 'save_dir': save_dir, 
        'recalculate': recalculate, 'rand_seed' : rng.integers(0, 2**32-1),
    }
    kwargs_components.update(kwargs)
    components = NMFcomponents(ref_psfs, **kwargs_components)

    bp_mask = np.isnan(sci)
    sci[bp_mask] = 0

    # Generate PSF models for each component
    ncomp_arr = np.arange(n_components) + 1
    model_arr = []
    for ncomp in ncomp_arr:
        kwargs_model = {
            'sci_err': sci_err, 'mask_data_imputation': mask_data_imputation, 
            'n_components': ncomp, 'oneByOne': False, 
            'maxiters': maxiters, 'tol': 1e-8, 'verbose': False,
            'rand_seed' : rng.integers(0, 2**32-1),
        }
        kwargs_model.update(kwargs)
        model = NMFmodelling(sci, components, **kwargs_model)
        model_arr.append(model)
    model_arr = np.array(model_arr)

    if mask_data_imputation is None:
        if ("planet" in sci_type) or (sci_type=="p"):
            # best_frac = 1
            best_frac_arr = np.ones(len(model_arr))
        elif ("disk" in sci_type) or (sci_type=="d"):
            # Find best scale factor for subtraction
            # best_frac = NMFbff(sci, model)
            best_frac_arr = np.array([NMFbff(sci, model, **kwargs) for model in model_arr]).flatten()
        
        result_arr = sci - model_arr * best_frac_arr.reshape([-1,1])
    else:                                
        result_arr = sci - model_arr

    result_arr[:, bp_mask] = np.nan

    return result_arr


def build_nmf_dict(fwhm_pix, **kwargs):
    """ Build PCA parameter dictionary

    Returns a dictionary of settings to pass to various analysis functions.
    There are multiple keywords that will be set by default and can be modified 
    by passing via keyword arguments. 
    
    Example output:

    nmf_params = {
        'fwhm_pix'    : fwhm_pix, # Number of pixels equal to PSF FWHM
        'loci'        : True,   # Locally optimized regions?
        'numbasis'    : [1,5,10,50],  # Number of KL basis vectors to use

        'kwargs_sub' : {}, # Subtraction regions config
        'kwargs_opt' : {}, # Optimization regions config
    }


    # Subtraction regions config
    kwargs_sub = {
        'IWA_nfwhm' : 0,     # Inner working angle in units of FWHM
        'OWA_nfwhm' : None,  # Outer working angle in units of FWHM
        'sub_ann_rad'   : 3, # Radial width of annulus (units of FWHM)
        'sub_ann_width' : 3, # Arc length of annulus (units of FWHM)
        'annuli_spacing' : 'constant', # Constant, log, or linear
        'constant_theta' : False,      # Do we want constant angle for each ring?
    }    

    # Optimization regions config (only for LOCI)
    kwargs_opt = {
        'opt_ann_rad'   : None, # Radial width of annulus (units of FWHM)
        'opt_ann_width' : 4,    # Arc length of annulus (units of FWHM)
        'nfwhm_sep'   : 0,      # Separation between subtraction and optimization regions (units of FWHM)
        'exclude_sub' : True,   # Exclude the subtraction region from the optimization region?
    }

    Parameters
    ----------
    param_dict : dict
        Dictionary of parameters from the main reduction script.
    """


    # PCA Parameters
    pca_params = {
        'fwhm_pix' : fwhm_pix, # Number of pixels equal to PSF FWHM

        # 'bin' : 1,               # Image binning factor
        # 'navg_ints_sci' : None,  # Number of sci INTS to average together
        # 'navg_ints_ref' : None,  # Number of ref INTS to average together
        # 'scale_psfs'    : False, # Scale PSFs by flux before subtracting and combining?
        # 'sat_rad_asec'  : None,  # Saturation radius in arcsec

        # 'add_fake' : False,    # Add Fake Planets?

        'loci'        : True,  # Use LOCI instead of KLIP?
        'remove_mean' : False,  # Remove mean from each frame before PCA?
        'numpsfs_max' : 100,    # Maximum number of PSFs to use for decomposition
        'numbasis'    : [1,5,10,50],  # Number of KL basis vectors to use
        'svd_frac'    : 0.9, # Fraction of variance to describe with KL basis vectors
    }

    # Subtraction regions config
    kwargs_sub = {
        'IWA_nfwhm' : 0,
        'OWA_nfwhm' : None,
        'sub_ann_rad'   : 3, # In terms of FWHM
        'sub_ann_width' : 3, # In terms of FWHM
        'annuli_spacing' : 'constant',
        'constant_theta' : False,
    }
    for k in kwargs_sub.keys():
        kwargs_sub[k] = kwargs.pop(k, kwargs_sub[k])

    # Optimization regions config
    kwargs_opt = {
        'opt_ann_rad'   : None,
        'opt_ann_width' : 2,
        'nfwhm_sep'   : 0,
        'exclude_sub' : True,
    }
    for k in kwargs_opt.keys():
        kwargs_opt[k] = kwargs.pop(k, kwargs_opt[k])
    # Default the radius of the optimization region to be 1 FWHM larger than the subtraction region
    if kwargs_opt['opt_ann_rad'] is None:
        kwargs_opt['opt_ann_rad'] = kwargs_sub['sub_ann_rad']+1

    # Save to dictionary
    pca_params['kwargs_sub'] = kwargs_sub
    pca_params['kwargs_opt'] = kwargs_opt

    # Update defaults with any user inputs
    pca_params.update(**kwargs)

    return pca_params


def NMFcomponents_old(ref, ref_err = None, n_components = None, maxiters = 1e3, oneByOne = False, ignore_mask = None, path_save = None, recalculate = False):
    """Returns the NMF components, where the rows contain the information.
    Args:
        ref and ref_err should be (N, p) where N is the number of references, p is the number of pixels in each reference.
        ignore_mask: array of shape (N, p). mask pixels in each image that you don't want to use. 
        path_save: string, path to save the NMF components (at: path_save + '_comp.fits') and coeffieients (at: path_save + '_coef.fits')
        recalculate: boolean, whether to recalculate when path_save is provided
    Returns: 
        NMf components (n_components * p).
    """
    ref = ref.T # matrix transpose to comply with statistician standards on storing data
    
    if ref_err is None:
        ref_err = np.sqrt(ref)
    else:
        ref_err = ref_err.T # matrix transpose for the error map as well
        
    if (n_components is None) or (n_components > ref.shape[0]):
        n_components = ref.shape[0]
        
    if ignore_mask is None:
        ignore_mask = np.ones_like(ref)
    
    # ignore certain values in component construction
    ignore_mask[ref <= 0] = 0 # 1. negative values
    ignore_mask[~np.isfinite(ref)] = 0 # 2. infinite values
    ignore_mask[np.isnan(ref)] = 0 # 3. nan values
    
    ignore_mask[ref_err <= 0] = 0 # 1. negative values in input error map
    ignore_mask[~np.isfinite(ref_err)] = 0 # 2. infinite values in input error map
    ignore_mask[np.isnan(ref_err)] = 0 # 3. nan values in input error map
    
    # speed up component calculation by ignoring the commonly-ignored elements across all references
    mask_mark = np.nansum(ignore_mask, axis = 1)
    mask_mark[mask_mark != 0] = 1 # 1 means that there is coverage in at least one of the refs
    
    ref_columnized = data_masked_only(ref, mask = mask_mark)
    ref_err_columnized = data_masked_only(ref_err, mask = mask_mark)
    mask_columnized = data_masked_only(ignore_mask, mask = mask_mark)
    mask_columnized_boolean = np.array(data_masked_only(ignore_mask, mask = mask_mark), dtype = bool)
    ref_columnized[mask_columnized == 0] = 0 # assign 0 to ignored values, should not impact the final result given the usage of mask_columnized_boolean
    ref_err_columnized[mask_columnized == 0] = np.nanmax(ref_err_columnized) # assign max uncertainty to ignored values, should not impact the final result

    
    # component calculation
    components_column = 0
    if not oneByOne:
        g_img = NMF(ref_columnized, V=1.0/ref_err_columnized**2, M = mask_columnized_boolean, n_components=n_components)
        chi2, time_used = g_img.SolveNMF(maxiters=maxiters)
        components_column = g_img.W/np.sqrt(np.nansum(g_img.W**2, axis = 0)) #normalize the components        
        components = data_masked_only_revert(components_column, mask = mask_mark)        
    else:
        print("Building components one-by-one...")
        if path_save is None or recalculate:
            if recalculate:
                print('Recalculating no matter if you have saved previous ones.')
            for i in range(n_components):
                print("\t" + str(i+1) + " of " + str(n_components))
                n = i + 1
                if (i == 0):
                    g_img = NMF(ref_columnized, V = 1.0/ref_err_columnized**2, M = mask_columnized_boolean, n_components= n)
                else:
                    W_ini = np.random.rand(ref_columnized.shape[0], n)
                    W_ini[:, :(n-1)] = np.copy(g_img.W)
                    W_ini = np.array(W_ini, order = 'F') #Fortran ordering, column elements contiguous in memory.
                
                    H_ini = np.random.rand(n, ref_columnized.shape[1])
                    H_ini[:(n-1), :] = np.copy(g_img.H)
                    H_ini = np.array(H_ini, order = 'C') #C ordering, row elements contiguous in memory.
                
                    g_img = NMF(ref_columnized, V = 1.0/ref_err_columnized**2, M = mask_columnized_boolean, W = W_ini, H = H_ini, n_components= n)
                chi2 = g_img.SolveNMF(maxiters=maxiters)
            
                components_column = g_img.W/np.sqrt(np.nansum(g_img.W**2, axis = 0)) #normalize the components
                components = data_masked_only_revert(components_column, mask = mask_mark) 
            if recalculate:
                print('\t\t\t Calculation for ' + str(n) + ' components done, overwriting raw 2D component matrix at ' + path_save + '_comp.fits')
                fits.writeto(path_save + '_comp.fits', g_img.W, overwrite = True)
                print('\t\t\t Calculation for ' + str(n) + ' components done, overwriting raw 2D coefficient matrix at ' + path_save + '_coef.fits')
                fits.writeto(path_save + '_coef.fits', g_img.H, overwrite = True)
        else:
            if not os.path.exists(path_save + '_comp.fits'):
                print('\t\t ' + path_save + '_comp.fits does not exist, calculating from scratch.')
                for i in range(n_components):
                    print("\t" + str(i+1) + " of " + str(n_components))
                    n = i + 1
                    if (i == 0):
                        g_img = NMF(ref_columnized, V = 1.0/ref_err_columnized**2, M = mask_columnized_boolean, n_components= n)
                    else:
                        W_ini = np.random.rand(ref_columnized.shape[0], n)
                        W_ini[:, :(n-1)] = np.copy(g_img.W)
                        W_ini = np.array(W_ini, order = 'F') #Fortran ordering, column elements contiguous in memory.
                
                        H_ini = np.random.rand(n, ref_columnized.shape[1])
                        H_ini[:(n-1), :] = np.copy(g_img.H)
                        H_ini = np.array(H_ini, order = 'C') #C ordering, row elements contiguous in memory.
                
                        g_img = NMF(ref_columnized, V = 1.0/ref_err_columnized**2, M = mask_columnized_boolean, W = W_ini, H = H_ini, n_components= n)
                    chi2 = g_img.SolveNMF(maxiters=maxiters)
                    print('\t\t\t Calculation for ' + str(n) + ' components done, overwriting raw 2D component matrix at ' + path_save + '_comp.fits')
                    fits.writeto(path_save + '_comp.fits', g_img.W, overwrite = True)
                    print('\t\t\t Calculation for ' + str(n) + ' components done, overwriting raw 2D coefficient matrix at ' + path_save + '_coef.fits')
                    fits.writeto(path_save + '_coef.fits', g_img.H, overwrite = True)
                    components_column = g_img.W/np.sqrt(np.nansum(g_img.W**2, axis = 0)) #normalize the components
                    components = data_masked_only_revert(components_column, mask = mask_mark)
            else:
                W_assign = fits.getdata(path_save + '_comp.fits')
                H_assign = fits.getdata(path_save + '_coef.fits')
                if W_assign.shape[1] >= n_components:
                    print('You have already had ' + str(W_assign.shape[1]) + ' components while asking for ' + str(n_components) + '. Returning to your input.')
                    components_column = W_assign/np.sqrt(np.nansum(W_assign**2, axis = 0))
                    components = data_masked_only_revert(components_column, mask = mask_mark)
                else:
                    print('You are asking for ' + str(n_components) + ' components. Building the rest based on the ' + str(W_assign.shape[1]) + ' provided.')

                    for i in range(W_assign.shape[1], n_components):
                        print("\t" + str(i+1) + " of " + str(n_components))
                        n = i + 1
                        if (i == W_assign.shape[1]):
                            W_ini = np.random.rand(ref_columnized.shape[0], n)
                            W_ini[:, :(n-1)] = np.copy(W_assign)
                            W_ini = np.array(W_ini, order = 'F') #Fortran ordering, column elements contiguous in memory.
            
                            H_ini = np.random.rand(n, ref_columnized.shape[1])
                            H_ini[:(n-1), :] = np.copy(H_assign)
                            H_ini = np.array(H_ini, order = 'C') #C ordering, row elements contiguous in memory.
            
                            g_img = NMF(ref_columnized, V = 1.0/ref_err_columnized**2, W = W_ini, H = H_ini, M = mask_columnized_boolean, n_components= n)
                        else:
                            W_ini = np.random.rand(ref_columnized.shape[0], n)
                            W_ini[:, :(n-1)] = np.copy(g_img.W)
                            W_ini = np.array(W_ini, order = 'F') #Fortran ordering, column elements contiguous in memory.
            
                            H_ini = np.random.rand(n, ref_columnized.shape[1])
                            H_ini[:(n-1), :] = np.copy(g_img.H)
                            H_ini = np.array(H_ini, order = 'C') #C ordering, row elements contiguous in memory.
            
                            g_img = NMF(ref_columnized, V = 1.0/ref_err_columnized**2, W = W_ini, H = H_ini, M = mask_columnized_boolean, n_components= n)
                        chi2 = g_img.SolveNMF(maxiters=maxiters)
                        print('\t\t\t Calculation for ' + str(n) + ' components done, overwriting raw 2D component matrix at ' + path_save + '_comp.fits')
                        fits.writeto(path_save + '_comp.fits', g_img.W, overwrite = True)
                        print('\t\t\t Calculation for ' + str(n) + ' components done, overwriting raw 2D coefficient matrix at ' + path_save + '_coef.fits')
                        fits.writeto(path_save + '_coef.fits', g_img.H, overwrite = True)
                        components_column = g_img.W/np.sqrt(np.nansum(g_img.W**2, axis = 0)) #normalize the components
                        components = data_masked_only_revert(components_column, mask = mask_mark)            
    return components.T


def NMFmodelling_old(trg, components, n_components = None, mask_components = None, mask_data_imputation = None, trg_err = None, maxiters = 1e3, cube = False, trgThresh = 0):
    """ NMF modeling.
    Args:
        trg: 1D array, p pixels
        components: N * p, calculated using NMFcomponents.
        n_components: how many components do you want to use. If None, all the components will be used.
        cube: whether output a cube or not (increasing the number of components).
        trgThresh: ignore the regions with low photon counts. Especially when they are ~10^-15 or smaller. I chose 0 in this case.
    
    Returns:
        NMF model of the target.
    """

    
    if n_components is None:
        n_components = components.shape[0]
        
    if trg_err is None:
        trg_err = np.sqrt(trg)

    if mask_components is None:
        mask_components = np.ones(trg.shape)
        mask_components[np.where(np.isnan(components[0]))] = 0

    components_column_all = data_masked_only(components[:n_components].T, mask = mask_components)   #columnize the components, make sure NonnegMFPy returns correct results.
    components_column_all = components_column_all/np.sqrt(np.nansum(components_column_all**2, axis = 0)) #normalize the components #make sure the components are normalized.
    
    if mask_data_imputation is None:
        flag_di = 0
        mask_data_imputation = np.ones(trg.shape)
    else:
        flag_di = 1
        print('Data Imputation!')
        
    mask = mask_components*mask_data_imputation #will be used for modeling

    trg[trg < trgThresh] = 0
    trg_err[trg == 0] = np.nanmax(trg_err)

    mask[trg <= 0] = 0
    mask[np.isnan(trg)] = 0
    mask[~np.isfinite(trg)] = 0
    
    #Columnize the target and its error.
    trg_column = data_masked_only(trg, mask = mask)
    trg_err_column = data_masked_only(trg_err, mask = mask)
    components_column = data_masked_only(components.T, mask = mask)

    if not cube:
        trg_img = NMF(trg_column, V=1/trg_err_column**2, W=components_column, n_components = n_components)
        (chi2, time_used) = trg_img.SolveNMF(H_only=True, maxiters = maxiters)
        coefs = trg_img.H
        if flag_di == 0: # do not do data imputation
            model_column = np.dot(components_column, coefs)

            model = data_masked_only_revert(model_column, mask)
            model[np.where(mask == 0)] = np.nan
        elif flag_di == 1: # do data imputation
            model_column = np.dot(components_column_all, coefs)
            model = data_masked_only_revert(model_column, mask_components)
            model[np.where(mask_components == 0)] = np.nan
    else:
        print("Building models one-by-one...")

        for i in range(n_components):
            print("\t" + str(i+1) + " of " + str(n_components))
            trg_img = NMF(trg_column, V=1/trg_err_column**2, W=components_column[:, :i+1], n_components = i + 1)
            (chi2, time_used) = trg_img.SolveNMF(H_only=True, maxiters = maxiters)

            coefs = trg_img.H

            model_column = np.dot(components_column[:, :i+1], coefs)

    return model.flatten() #model_column.T.flatten()


# from .nrc_utils import *
import numpy as np
from .nrc_utils import stpsf, poppy, read_filter
from .logging_utils import setup_logging
from .pynrc_core import NIRCam
from .opds import opd_default

def perform_benchmarks(filter='F430M', pupil=None, mask=None, module='A',
                       fov_pix=33, oversample=4, include_si_wfe=True, 
                       include_ote_field_dependence=True, include_distortions=True,
                       use_legendre=True, force=False, save=True, 
                       do_stpsf=True, do_stpsf_only=False, return_nrc=False,
                       use_mp=None, nproc=None, **kwargs):
    
    import datetime, time
    
    # PSF setup
#     kwargs = {}
    kwargs['pupil'] = 'CLEAR' if pupil is None else pupil
    kwargs['mask'] = mask
    kwargs['module'] = module

    kwargs['fov_pix'] = fov_pix
    kwargs['oversample'] = oversample

    # kwargs['opd'] = ('OPD_RevW_ote_for_NIRCam_predicted.fits.gz', 0)
    kwargs['opd'] = opd_default
    kwargs['jitter'] = 'gaussian'
    kwargs['jitter_sigma'] = 0.007

    kwargs['include_si_wfe'] = include_si_wfe
    kwargs['include_ote_field_dependence'] = include_ote_field_dependence
    kwargs['include_distortions'] = include_distortions
    kwargs['use_legendre'] = use_legendre

    kwargs['force']     = force
    kwargs['save']      = save
    kwargs['save_name'] = None
    
    setup_logging('WARN', verbose=False)
    
    tdict = {
        'stpsf_init': None,
        'stpsf_psf': None,
        'pynrc_coeff': None,
        'pynrc_drift': None,
        'pynrc_field': None,
        'pynrc_psf': None,
    }
    
    # Multiprocessing cases
    use_mp_def = poppy.conf.use_multiprocessing
    nproc_def = poppy.conf.n_processes
    if use_mp is not None:
        poppy.conf.use_multiprocessing = use_mp
    if nproc is not None:
        poppy.conf.n_processes = nproc

    ####
    # STPSF timings
    ####
    if do_stpsf or do_stpsf_only:
        # Initialization overheads
        tarr = []
        for i in range(5):
            t0 = time.time()
            inst = stpsf.NIRCam()
            inst.detector = 'NRCA5'
            t1 = time.time()
            tarr.append(t1-t0)
        tdict['stpsf_init'] = dt = np.mean(tarr)
        time_string = 'Took {:.2f} seconds to init STPSF'.format(dt)
        print(time_string)    

        # PSF Generation
        nrc = NIRCam(filter=filter, **kwargs)
        tarr = []
        for i in range(5):
            t0 = time.time()
            _ = nrc.calc_psf()
            t1 = time.time()
            dt = t1-t0
            tarr.append(dt)
        tdict['stpsf_psf'] = dt = np.mean(tarr)
        time_string = 'Took {:.2f} seconds to generate STPSF PSF'.format(dt)
        print(time_string)    

    if do_stpsf_only and (not return_nrc):
        return tdict
    
    ####
    # pyNRC coefficients
    ####
    t0 = time.time()
    nrc = NIRCam(filter=filter, **kwargs)
    t1 = time.time()
    
    if do_stpsf_only and return_nrc:
        return nrc

    tdict['pynrc_coeff'] = dt = t1-t0
    time_string = 'Took {:.2f} seconds to generate pyNRC coefficients'.format(dt)
    print(time_string)

    ####
    # WFE drift coefficient generation
    ####
    t0 = time.time()
    nrc.gen_wfedrift_coeff()
    t1 = time.time()
    dt = t1-t0

    tdict['pynrc_drift'] = dt = t1-t0
    time_string = 'Took {:.2f} seconds to generate WFE Drift coefficients'.format(dt)
    print(time_string)

    ####
    # Field coefficient generation
    ####
    t0 = time.time()
    nrc.gen_wfefield_coeff()
    t1 = time.time()

    tdict['pynrc_field'] = dt = t1-t0
    time_string = 'Took {:.2f} seconds to generate WFE Field coefficients'.format(dt)
    print(time_string)
    
    tarr = []
    for i in range(10):
        t0 = time.time()
        
        _ = nrc.calc_psf_from_coeff(wfe_drift=5, coord_vals=(1024,1024), coord_frame='sci', 
                                    return_oversample=True, return_hdul=False)
        t1 = time.time()
        dt = t1-t0
        tarr.append(dt)
        if dt>10:
            break
    tdict['pynrc_psf'] = dt = np.mean(tarr)
    time_string = 'Took {:.2f} seconds to generate pynrc PSF'.format(dt)
    print(time_string)
    
    # Return defaults
    poppy.conf.use_multiprocessing = use_mp_def
    poppy.conf.n_processes = nproc_def
    
    if return_nrc:
        return nrc
    else:
        return tdict



"""
NGHXRG - Teledyne HxRG Noise Generator

**DEPRECATED FOR ngNRC.py**

Modification History:

8-15 April 2015, B.J. Rauscher, NASA/GSFC

- Implement Pierre Ferruit's (ESA/ESTEC)recommendation to use
  numpy.fft.rfft and numpy.fft.irfft for faster execution. This saves
  about 30% in execution time.
- Clean up setting default values per several suggestions from
  Chaz Shapiro (NASA/JPL).
- Change from setting the shell variable H2RG_PCA0 to point to the PCA-zero
  file to having a shell variable point to the NG home directory. This is per
  a suggestion from Chaz Shapiro (NASA/JPL) that would allow seamlessly adding  
  more PCA-zero templates.
- Implement a request form Chaz Shapiro (NASA/JPL) for status
  reporting. This was done by adding the "verbose" arguement.
- Implement a request from Pierre Ferruit (ESA/ESTEC) to generate
  3-dimensional data cubes.
- Implement a request from Pierre Ferruit to treat ACN as different 1/f
  noise in even/odd columns. Previously ACN was treated purely as a feature
  in Fourier space.
- Version 2(Beta)

16 April 2015, B.J. Rauscher

- Fixed a bug in the pinkening filter definitions. Abs() was used where
  sqrt() was intended. The bug caused power spectra to have the wrong shape at
  low frequency.
- Version 2.1(Beta)

17 April 2015, B.J. Rauscher

- Implement a request from Chaz Shapiro for HXRGNoise() to exit gracefully if
  the bias_file is not found.
- Version 2.2 (Beta)

8 July 2015, B.J. Rauscher

- Address PASP referee comments
    * Fast scan direction is now reversible. To reverse the slow scan
      direction use the numpy flipud() function.
    * Modifications to support subarrays. Specifically,
        > Setting reference_pixel_border_width=0 (integer zero);
            + (1) eliminates the reference pixel border and
            + (2) turns off adding in a bias pattern
                  when simulating data cubes. (Turned on in v2.5)
- Version 2.4

12 Oct 2015, J.M. Leisenring, UA/Steward

- Make compatible with Python 2.x
    * from __future__ import division
- Allow pca0 & kTC noise to show up for a single frame
- Included options for subarray modes (FULL, WINDOW, and STRIPE)
    * Keywords x0 and y0 define a subarray position (lower left corner)
    * Selects correct pca0 region for subarray underlay
    * Adds reference pixels if they exist within subarray window
- Tie negative values to 0 and anything >=2^16 to 2^16-1
- Version 2.5

20 Oct 2015, J.M. Leisenring, UA/Steward

- Padded nstep to the next power of 2 in order to improve FFT runtime
    * nstep2 = int(2**np.ceil(np.log2(nstep)))
    * Speeds up FFT calculations by ~5x
- Don't generate noise elements if their magnitudes are equal to 0.
- Returns a copy of final HDU result for easy retrieval
- Version 2.6

17 Nov 2015, J.M. Leisenring, UA/Steward

- Read in bias_file rather than pca0_file
    * Calculate pca0 element by normalizing bias file
- Include alternating column offsets (aco_a & aco_b)
    * These can be defined as a list or np array
- Version 2.7

15 Feb 2016, J.M. Leisenring, UA/Steward

- Add a reference instability
    * ref pixels don't track active pixels perfectly
- Version 2.8

9 May 2016, J.M. Leisenring, UA/Steward

- Each channel can have it's own read noise value
- Version 2.9

21 July 2016, J.M. Leisenring, UA/Steward

- Add option to use FFTW
    * This can be faster for some processors/OS architectures
    * The decision to use this is highly dependent on the computer
      and should be tested beforehand.
    * For more info: https://pypi.python.org/pypi/pyFFTW
- Version 3.0

"""
# Necessary for Python 2.6 and later
#from __future__ import division, print_function
from __future__ import absolute_import, division, print_function, unicode_literals

__version__ = "3.0"

import os
import warnings
from astropy.io import fits
import numpy as np
from scipy.ndimage.interpolation import zoom
from scipy.ndimage import convolve
from astropy.stats.funcs import median_absolute_deviation as mad
import datetime
# import matplotlib.pyplot as plt # Handy for debugging

try:
    import pyfftw
    pyfftw_available = True
except ImportError:
    pyfftw_available = False
import multiprocessing as mp
#import time


warnings.filterwarnings('ignore')
import logging
_log = logging.getLogger('nghxrg')

class HXRGNoise:
    """Simulate Teledyne HxRG + SIDECAR ASIC noise
    
    HXRGNoise is a class for making realistic Teledyne HxRG system
    noise. The noise model includes correlated, uncorrelated,
    stationary, and non-stationary components. The default parameters
    make noise that resembles Channel 1 of JWST NIRSpec. NIRSpec uses
    H2RG detectors. They are read out using four video outputs at
    1.e+5 pix/s/output.
    
    Parameters
    ----------
    naxis1 : int
        X-dimension of the FITS cube.
    naxis2 : int
        Y-dimension of the FITS cube.
    naxis3 : int
        Z-dimension of the FITS cube (number of up-the-ramp samples).
    n_out : int
        Number of detector amplifiers/channels/outputs.
    nroh : int
        New row overhead in pixels. This allows for a short
        wait at the end of a row before starting the next one.
    nfoh : int
        New frame overhead in rows. This allows for a short
        wait at the end of a frame before starting the next one.
    nfoh_pix(TBD) : int
        New frame overhead in pixels. This allows for a short
        wait at the end of a frame before starting the next one.
        Generally a single pix offset for full frame and stripe
        for JWST ASIC systems.
    dt : float
        Pixel dwell time in seconds (10e-6 sec, for instance).
    bias_file : str
        Name of a FITS file that contains bias pattern, also used for PCA-zero.
    dark_file : str
        Name of a FITS file that contains dark current values per pixel.
    verbose : bool
        Enable this to provide status reporting.
    wind_mode : str
        'FULL', 'STRIPE', or 'WINDOW'.
    x0/y0 : int
        Pixel positions of subarray mode.
    det_size : int
        Pixel dimension of full detector (square).
    reference_pixel_border_width : int
        Width of reference pixel border around image area.
    same_scan_direction : bool
        Are all the output channels read in the same direction?
        By default fast-scan readout direction is ``[-->,<--,-->,<--]``
        If ``same_scan_direction``, then all ``-->``
    reverse_scan_direction : bool
        Enable this to reverse the fast scanner readout directions. 
        This capability was added to support Teledyne's programmable 
        fast scan readout directions. The default setting of False 
        corresponds to what HxRG detectors default to upon power up.
        If ``reverse_scan_direction``, then ``[<--,-->,<--,-->]`` or all ``<--``
    use_fftw : bool
        If pyFFTW is installed, you can use this in place of np.fft.
    ncores : int
        Specify number of cores (threads, actually) to use for pyFFTW.
    
    """

    # These class variables are common to all HxRG detectors
    nghxrg_version = float(__version__) # Sofware version

    def __init__(self, naxis1=None, naxis2=None, naxis3=None, n_out=None,
                 dt=None, nroh=None, nfoh=None, nfoh_pix=None,
                 dark_file=None, bias_file=None, verbose=False,
                 reverse_scan_direction=False, same_scan_direction=False,
                 reference_pixel_border_width=None,
                 wind_mode='FULL', x0=0, y0=0, det_size=None, 
                 use_fftw=False, ncores=None):

        # pyFFTW usage
        self.use_fftw = True if (use_fftw and pyfftw_available) else False
        # By default, use 50% of available cores for FFTW parallelization
        self.ncores = mp.cpu_count() // 2 if ncores is None else int(ncores)	

        # ======================================================================
        #
        # DEFAULT CLOCKING PARAMETERS
        #
        # The following parameters define the default HxRG clocking pattern. The
        # parameters that define the default noise model are defined in the
        # mknoise() method.
        #
        # ======================================================================

        # Subarray?
        if wind_mode is None:
            wind_mode = 'FULL'
        if det_size is None:
            det_size = 2048
        wind_mode = wind_mode.upper()
        modes = ['FULL', 'STRIPE', 'WINDOW']
        if wind_mode not in modes:
            _log.warning(f'{wind_mode} not a valid window readout mode! Returning...')
            os.sys.exit()
        if wind_mode == 'WINDOW':
            n_out = 1
        if wind_mode == 'FULL':
            x0 = 0; y0 = 0
        if wind_mode == 'STRIPE':
            x0 = 0

        # Default clocking pattern is JWST NIRSpec
        self.naxis1    = 2048  if naxis1   is None else int(naxis1)
        self.naxis2    = 2048  if naxis2   is None else int(naxis2)
        self.naxis3    = 1     if naxis3   is None else int(naxis3)
        self.n_out     = 4     if n_out    is None else int(n_out)
        self.dt        = 10e-6 if dt       is None else dt
        self.nroh      = 12    if nroh     is None else int(nroh)
        self.nfoh      = 1     if nfoh     is None else int(nfoh)
        self.nfoh_pix  = 0     #if nfoh_pix is None else int(nfoh_pix)
        self.reference_pixel_border_width = 4 if reference_pixel_border_width is None \
                                              else int(reference_pixel_border_width)
                                  
        # Check that det_size is greater than self.naxis1 and self.naxis2 in WINDOW mode (JML)
        if wind_mode == 'WINDOW':
            if (self.naxis1 > det_size):
                _log.warning('NAXIS1 %s greater than det_size %s! Returning...' % (self.naxis1,det_size))
                os.sys.exit()
            if (self.naxis2 > det_size):
                _log.warning('NAXIS2 %s greater than det_size %s! Returning...' % (self.naxis1,det_size))
                os.sys.exit()

        # Initialize PCA-zero file and make sure that it exists and is a file
        #self.bias_file = os.getenv('NGHXRG_HOME')+'/sca_images/nirspec_pca0.fits' if \
        #				 bias_file is None else bias_file
        #self.bias_file = 'nirspec_pca0.fits' if bias_file is None else bias_file
        self.bias_file = bias_file
        if bias_file is not None:
            if os.path.isfile(self.bias_file) is False:
                raise ValueError('There was an error finding bias_file {}'.format(bias_file))
                # print('There was an error finding bias_file!')
                # print(bias_file)
                #os.sys.exit()

#             print('There was an error finding bias_file! Check to be')
#             print('sure that the NGHXRG_HOME shell environment')
#             print('variable is set correctly and that the')
#             print('$NGHXRG_HOME/ directory contains the desired PCA0')
#             print('file. The default is nirspec_pca0.fits.')
#             os.sys.exit()


        # Add in dark current file (JML)
        self.dark_file = dark_file
        if dark_file is not None:
            if os.path.isfile(self.dark_file) is False:
                raise ValueError('There was an error finding dark_file {}'.format(dark_file))
                #print('There was an error finding dark_file!')
                #print(dark_file)
                #os.sys.exit()


        # ======================================================================

        # Configure Subarray
        self.wind_mode = wind_mode
        self.det_size  = det_size
        self.x0 = x0
        self.y0 = y0

        # Configure status reporting
        self.verbose = verbose

        # Configure readout direction
        self.reverse_scan_direction = reverse_scan_direction
        self.same_scan_direction = same_scan_direction


        # Compute the number of pixels in the fast-scan direction per output
        self.xsize = self.naxis1 // self.n_out
    
        # Compute the number of time steps per integration, per output
        self.nstep_frame = (self.xsize+self.nroh) * (self.naxis2+self.nfoh) + self.nfoh_pix
        self.nstep =  self.nstep_frame * self.naxis3
        # Pad nsteps to a power of 2, which is much faster
        self.nstep2 = int(2**np.ceil(np.log2(self.nstep)))

        # Compute frame time and ramp time
        self.tframe = self.nstep_frame * self.dt
        self.inttime = self.tframe * self.naxis3

        # For adding in ACN, it is handy to have masks of the even
        # and odd pixels on one output neglecting any gaps
        self.m_even = np.zeros((self.naxis3,self.naxis2,self.xsize))
        self.m_odd = np.zeros_like(self.m_even)
        for x in np.arange(0,self.xsize,2):
            self.m_even[:,:self.naxis2,x] = 1
            self.m_odd[:,:self.naxis2,x+1] = 1
        self.m_even = np.reshape(self.m_even, np.size(self.m_even))
        self.m_odd = np.reshape(self.m_odd, np.size(self.m_odd))

        # Also for adding in ACN, we need a mask that point to just
        # the real pixels in ordered vectors of just the even or odd
        # pixels
        self.m_short = np.zeros((self.naxis3, self.naxis2+self.nfoh, \
                                      (self.xsize+self.nroh)//2))
        self.m_short[:,:self.naxis2,:self.xsize//2] = 1
        self.m_short = np.reshape(self.m_short, np.size(self.m_short))

        # Define frequency arrays       
        self.f1 = np.fft.rfftfreq(self.nstep2) # Frequencies for nstep elements
        self.f2 = np.fft.rfftfreq(2*self.nstep2) # ... for 2*nstep elements
        self.f3 = np.fft.rfftfreq(2*self.naxis3)

        # First element should not be 0
        self.f1[0] = self.f1[1]
        self.f2[0] = self.f2[1]
        self.f3[0] = self.f3[1]

        # Define pinkening filters. F1 and p_filter1 are used to
        # generate ACN. F2 and p_filter2 are used to generate 1/f noise.
        # F2 and p_filter2 are used to generate reference instabilities.
        self.alpha = -1 # Hard code for 1/f noise until proven otherwise
        self.p_filter1 = np.sqrt(self.f1**self.alpha)
        self.p_filter2 = np.sqrt(self.f2**self.alpha)
        self.p_filter3 = np.sqrt(self.f3**self.alpha)
        self.p_filter1[0] = 0.
        self.p_filter2[0] = 0.
        self.p_filter3[0] = 0.


        # Initialize pca0. This includes scaling to the correct size,
        # zero offsetting, and renormalization. We use robust statistics
        # because pca0 is real data
        if self.bias_file is None:
            h = fits.PrimaryHDU(np.zeros([det_size, det_size]))
            hdu = fits.HDUList([h])
        else:
            hdu = fits.open(self.bias_file)
        nx_pca0 = hdu[0].header['naxis1']
        ny_pca0 = hdu[0].header['naxis2']
        data = hdu[0].data     

        # Make sure the real PCA image is correctly scaled to size of fake data (JML)
        # Depends if we're FULL, STRIPE, or WINDOW
        if wind_mode == 'FULL':
            scale1 = self.naxis1 / nx_pca0
            scale2 = self.naxis2 / ny_pca0
            zoom_factor = np.max([scale1, scale2])
        if wind_mode == 'STRIPE':
            zoom_factor = self.naxis1 / nx_pca0
        if wind_mode == 'WINDOW':
            # Scale based on det_size
            scale1 = self.det_size / nx_pca0
            scale2 = self.det_size / ny_pca0
            zoom_factor = np.max([scale1, scale2])

        # Resize PCA0 data
        #print(zoom_factor)
        if zoom_factor != 1:
            data = zoom(data, zoom_factor, order=1, mode='wrap')
    
        # Copy data to save as bias pattern
        bias_image = data.copy()

        # Renormalize for PCA0 noise stuff
        data -= np.median(data) # Zero offset
        data /= (1.4826 * mad(data)) # Renormalize

        # Select region of pca0 associated with window position
        if self.wind_mode == 'WINDOW':
            x1 = self.x0; y1 = self.y0
        elif self.wind_mode == 'STRIPE':
            x1 = 0; y1 = self.y0
        else:
            x1 = 0; y1 = 0

        x2 = x1 + self.naxis1
        y2 = y1 + self.naxis2
        # Make sure x2 and y2 are valid
        if (x2 > data.shape[0] or y2 > data.shape[1]):
            _log.warning('Specified window size does not fit within detector array!')
            _log.warning('X indices: [%s,%s]; Y indices: [%s,%s]; XY Size: [%s, %s]' % 
                        (x1,x2,y1,y2,data.shape[0],data.shape[1]))
            os.sys.exit()
    
        # Save as properties
        self.pca0 = data[y1:y2,x1:x2]
        self.bias_image = bias_image[y1:y2,x1:x2]

        # Open dark current file (ADU/sec/pixel)
        if self.dark_file is not None:
            dark_hdu = fits.open(self.dark_file)
            dark_image = dark_hdu[0].data
            self.dark_image = dark_image[y1:y2,x1:x2]
            # Dark current distributions are very wide because of uncertainties
            # This causes certain pixels to fall below 0.
            # We can assume all pixels within 5-sigma have the same dark current
            #   as well as those with negative values.
            # Those with large dark currents are likely real.
            sig = 1.4826 * mad(dark_image)
            med = np.median(dark_image)
            l1 = med - 5*sig; l2 = med + 5*sig
            self.dark_image[(self.dark_image > l1) & (self.dark_image < l2)] = med
            # Set negative values to median
            self.dark_image[self.dark_image<0] = med
    
            # Set negative values to median
            #self.dark_image[self.dark_image<0] = np.median(self.dark_image)
            #self.dark_image[self.dark_image<0.005] = 0.001
        else:
            self.dark_image = None

        # How many reference pixels on each border?
        w = self.reference_pixel_border_width # Easier to work with
        lower = w-y1; upper = w-(det_size-y2)
        left = w-x1; right = w-(det_size-x2)
        ref_all = np.array([lower,upper,left,right])
        ref_all[ref_all<0] = 0
        self.ref_all = ref_all


    def message(self, message_text):
        """Used for status reporting"""
        if self.verbose is True:
            print('NG: ' + message_text + ' at DATETIME = ', (datetime.datetime.now().time()))

    def white_noise(self, nstep=None):
        """Gaussian noise
        
        Generate white noise for an HxRG including all time steps
        (actual pixels and overheads).

        Parameters
        ----------
        nstep : int
            Length of vector returned
        """
        return(np.random.standard_normal(nstep))    

    def pink_noise(self, mode, fmin=None):
        """Generate a vector of non-periodic pink noise.

        Parameters
        ----------
        mode : str
            Selected from 'pink', 'acn', or 'ref_inst'.
        fmin : float, optional
            Low-frequency cutoff. A value of 0 means no cut-off.
        """

        # Configure depending on mode setting
        if 'pink' in mode:
            nstep  = 2*self.nstep
            nstep2 = 2*self.nstep2
            f = self.f2
            p_filter = self.p_filter2
        elif 'acn' in mode:
            nstep  = self.nstep
            nstep2 = self.nstep2
            f = self.f1
            p_filter = self.p_filter1
        elif 'ref_inst' in mode:
            nstep  = 2*self.naxis3
            nstep2 = 2*self.naxis3
            f = self.f3
            p_filter = self.p_filter3

        # Build scaling factors for all frequencies
        fmin = 1./nstep2 if fmin is None else np.max([fmin, 1./nstep2])
        ix  = np.sum(f < fmin)   # Index of the cutoff
        if ix > 1 and ix < len(f):
            f = f.copy()
            p_filter = p_filter.copy()
            f[:ix] = f[ix]
            p_filter[:ix] = p_filter[ix]

        # Calculate theoretical output standard deviation from scaling
        w = p_filter[1:-1]
        w_last = p_filter[-1] * (1 + (nstep2 % 2)) / 2. # correct f = +-0.5
        the_std = 2 * np.sqrt(np.sum(w**2) + w_last**2) / nstep2

        # Generate scaled random power + phase
        sr = np.random.normal(scale=p_filter)
        si = np.random.normal(scale=p_filter)

        # If the signal length is even, frequencies +/- 0.5 are equal
        # so the coefficient must be real.
        if (nstep2 % 2) == 0: 
            si[-1] = 0

        # Regardless of signal length, the DC component must be real
        si[0] = 0

        # Combine power + corrected phase to Fourier components
        thefft  = sr + 1J * si

        #p0 = time.time()
        # Apply the pinkening filter.
        if self.use_fftw:
            result = pyfftw.interfaces.numpy_fft.irfft(thefft, overwrite_input=True,\
                planner_effort='FFTW_ESTIMATE', threads=self.ncores)
        else:
            result = np.fft.irfft(thefft)

        #p1 = time.time()
        #print("FFT and IFFT took",p1-p0," seconds")

        # Keep 1st half of nstep and scale to unit variance
        result = result[:nstep//2] / the_std
  
        return(result)



    def mknoise(self, o_file=None, gain=None,
                rd_noise=None, c_pink=None, u_pink=None, 
                acn=None, aco_a=None, aco_b=None, pca0_amp=None,
                reference_pixel_noise_ratio=None, ktc_noise=None,
                bias_off_avg=None, bias_off_sig=None, bias_amp=None,
                ch_off=None, ref_f2f_corr=None, ref_f2f_ucorr=None, ref_inst=None,
                out_ADU=True):
        """Create FITS cube containing only noise

        Parameters
        ----------
        o_file : str, None
            Output filename. If None, then no output.
        gain : float
            Gain in e/ADU. Defaults to 1.0.
        ktc_noise : float
            kTC noise in electrons. Set this equal to 
            sqrt(k*T*C_pixel)/q_e, where k is Boltzmann's constant, 
            T is detector temperature, and C_pixel is pixel capacitance. 
            For an H2RG, the pixel capacitance is typically about 40 fF.
        rd_noise : float
            Standard deviation of read noise in electrons. 
            Can be an array for individual amplifiers. 
        c_pink :float
            Standard deviation of correlated pink noise in electrons.
        u_pink : float
            Standard deviation of uncorrelated pink noise in electrons.
            Can be an array for individual amplifiers. 
        acn : float
            Standard deviation of alterating column noise in electrons
        pca0_amp : float
            Standard deviation of pca0 in electrons
        reference_pixel_noise_ratio : float
            Ratio of the standard deviation of the reference pixels to 
            the science pixels. Reference pixels are usually a little 
            lower noise.                                          
        bias_off_avg : float
            On average, integrations start here in electrons. 
            Set this so that all pixels are in range.
        bias_off_sig : float
            bias_off_avg has some variation. This is its std dev.
        bias_amp : float
            A multiplicative factor that we multiply bias_image by
            to simulate a bias pattern. This is completely
            independent from adding in "picture frame" noise. Set to
            0.0 remove bias pattern. For NIRCam, default is 1.0.
        ch_off : float
            Offset of each channel relative to bias_off_avg.
            Can be an array for individual amplifiers. 
        ref_f2f_corr : float
            Random frame-to-frame reference offsets due to PA reset,
            correlated between channels.
        ref_f2f_ucorr : float
            Random frame-to-frame reference offsets due to PA reset,
            per channel. Can be an array for individual amplifiers.
        aco_a : float
            Relative offsets of altnernating columns "a".
            Can be an array for individual amplifiers.
        aco_b : float
            Relative offsets of altnernating columns "b".
            Can be an array for individual amplifiers.
        ref_inst : float
            Reference instability relative to active pixels.
        out_ADU : bool
            Return as converted to ADU (True) or raw electrons?
              
        Notes
        -----
        Because of the noise correlations, there is no simple way to
        predict the noise of the simulated images. However, to a
        crude first approximation, these components add in
        quadrature.

        The units in the above are mostly "electrons". This follows convention
        in the astronomical community. From a physics perspective, holes are
        actually the physical entity that is collected in Teledyne's p-on-n
        (p-type implants in n-type bulk) HgCdTe architecture.
        """

        self.message('Starting mknoise()')

        # ======================================================================
        #
        # DEFAULT NOISE PARAMETERS
        #
        # These defaults create noise similar to that seen in the JWST NIRSpec.
        #
        # ======================================================================

        #self.gain      = 1.0      if gain         is None else gain
        #self.rd_noise  = 5.2      if rd_noise     is None else rd_noise
        #self.c_pink    = 3.0      if c_pink       is None else c_pink
        #self.u_pink    = 1.0      if u_pink       is None else u_pink
        #self.acn       = 0.5      if acn          is None else acn
        #self.pca0_amp  = 0.2      if pca0_amp     is None else pca0_amp

        self.gain     = 1.0  if gain     is None else gain
        # Set certain values equal to None if they are set to 0.
        #self.rd_noise = None if rd_noise == 0.0  else rd_noise
        self.rd_noise = rd_noise
        self.c_pink   = None if c_pink   == 0.0  else c_pink
        self.u_pink   = u_pink
        self.acn      = None if acn      == 0.0  else acn
        self.pca0_amp = None if pca0_amp == 0.0  else pca0_amp


        # Change this only if you know that your detector is different from a
        # typical H2RG.
        self.reference_pixel_noise_ratio = 0.8 if \
            reference_pixel_noise_ratio is None else reference_pixel_noise_ratio

        # These are used only when generating cubes. They are
        # completely removed when the data are calibrated to
        # correlated double sampling or slope images. We include
        # them in here to make more realistic looking raw cubes.
        self.ktc_noise     = 29.    if ktc_noise     is None else ktc_noise 
        self.bias_off_avg  = 5000.  if bias_off_avg  is None else bias_off_avg
        self.bias_off_sig  = 0.     if bias_off_sig  is None else bias_off_sig
        self.bias_amp      = 1.     if bias_amp      is None else bias_amp
        self.ch_off        = 0.     if ch_off        is None else ch_off
        self.aco_a         = 0.     if aco_a         is None else aco_a
        self.aco_b         = 0.     if aco_b         is None else aco_b
        self.ref_f2f_corr  = None   if ref_f2f_corr  is None else ref_f2f_corr
        self.ref_f2f_ucorr = None   if ref_f2f_ucorr is None else ref_f2f_ucorr
        self.ref_inst      = None   if ref_inst      is None else ref_inst

        # ======================================================================

        # Initialize the result cube and add a bias pattern.
        self.message('Initializing results cube')
        result = np.zeros((self.naxis3, self.naxis2, self.naxis1), dtype=np.float32)
                  
        # Inject a bias pattern.
        bias_pattern = self.bias_image*self.bias_amp

        # Add overall bias offset plus random component
        bias_pattern += self.bias_off_avg + self.bias_off_sig * np.random.randn()	
            
        # Add in some kTC noise. Since this should always come out
        # in calibration, we do not attempt to model it in detail.
        if self.ktc_noise > 0:
            bias_pattern += self.ktc_noise * np.random.standard_normal((self.naxis2, self.naxis1))

        # Add pedestal offset to each output channel
        # Check if self.ch_off is a numpy array or list
        if isinstance(self.ch_off, (np.ndarray,list)):
            temp = np.asarray(self.ch_off)
            if temp.size != self.n_out:
                _log.warning('Number of elements in ch_off not equal to n_out')
                os.sys.exit()
            for ch in range(self.n_out):
                bias_pattern[:,self.xsize*ch:self.xsize*(ch+1)] += temp[ch]
        else:
            bias_pattern += self.ch_off

        # Add in alternating column offsets to bias pattern
        inda = np.arange(0,self.xsize,2)
        indb = inda+1		
        # Check if self.aco_a/b are numpy arrays or lists
        if isinstance(self.aco_a, (np.ndarray,list)):
            temp = np.asarray(self.aco_a) # Always set to a numpy array
            if temp.size != self.n_out:
                _log.warning('Number of elements in aco_a not equal to n_out')
                os.sys.exit()
        else: # Assumes aco_a is a single value as opposed to an array or list
            temp = np.ones(self.n_out) * self.aco_a
        # Add alternating column offsets for each channel
        for ch in range(self.n_out):
            chan = bias_pattern[:,self.xsize*ch:self.xsize*(ch+1)]
            chan[:,inda] += temp[ch]
        
        # Do the same, but with column b
        if isinstance(self.aco_b, (np.ndarray,list)):
            temp = np.asarray(self.aco_b) # Always set to a numpy array
            if temp.size != self.n_out:
                _log.warning('Number of elements in aco_b not equal to n_out')
                os.sys.exit()
        else: # Assumes aco_b is a single value as opposed to an array or list
            temp = np.ones(self.n_out) * self.aco_b
        # Add alternating column offsets for each channel
        for ch in range(self.n_out):
            chan = bias_pattern[:,self.xsize*ch:self.xsize*(ch+1)]
            chan[:,indb] += temp[ch]
    
        # Add in the bias pattern
        for z in np.arange(self.naxis3):
            result[z,:,:] += bias_pattern

        # Add in random frame-to-frame bias offsets
        # First, correlated bias between channels
        if self.ref_f2f_corr is not None:
            for z in np.arange(self.naxis3):
                result[z,:,:] += self.ref_f2f_corr * np.random.randn()
        # Next, channel-specific bias offsets
        if self.ref_f2f_ucorr is not None:
            if isinstance(self.ref_f2f_ucorr, (np.ndarray,list)):
                temp = np.asarray(self.ref_f2f_ucorr)
                if temp.size != self.n_out:
                    _log.warning('Number of elements in ref_f2f_ucorr not equal to n_out')
                    os.sys.exit()
            else: # Single value as opposed to an array or list
                temp = np.ones(self.n_out) * self.ref_f2f_ucorr
            for z in np.arange(self.naxis3):
                for ch in range(self.n_out):
                    result[z,:,self.xsize*ch:self.xsize*(ch+1)] += temp[ch] * np.random.randn()
        # Reference instability (frame-to-frame reference offset not recorded in active pixels)
        if self.ref_inst is not None:
            ref_noise = self.ref_inst * self.pink_noise('ref_inst')
            w = self.ref_all
            for z in np.arange(self.naxis3):
                if w[0] > 0:
                    result[z, :w[0], :] += ref_noise[z]
                if w[1] > 0:
                    result[z, -w[1]:,:] += ref_noise[z]
                if w[2] > 0:
                    result[z, :, :w[2]] += ref_noise[z]
                if w[3] > 0:
                    result[z, :,-w[3]:] += ref_noise[z]
            

        # Make white read noise. This is the same for all pixels.
        if self.rd_noise is not None:
            self.message('Generating rd_noise')

            # We want self.rd_noise to be an array or list
            if isinstance(self.rd_noise, (np.ndarray,list)):
                temp = np.asarray(self.rd_noise)
                if temp.size != self.n_out:
                    _log.warning('Number of elements in rd_noise not equal to n_out')
                    os.sys.exit()
            else: # Single value as opposed to an array or list
                self.rd_noise = np.ones(self.n_out) * self.rd_noise
    
            w = self.ref_all
            r = self.reference_pixel_noise_ratio  # Easier to work with
            for z in np.arange(self.naxis3):
                here = np.zeros((self.naxis2, self.naxis1))
        
                # First assume no ref pixels and just add in random noise
                for op in np.arange(self.n_out):
                    x0 = op * self.xsize
                    x1 = x0 + self.xsize
                    here[:,x0:x1] = self.rd_noise[op] * np.random.standard_normal((self.naxis2,self.xsize))
        
                # If there are reference pixels, overwrite with appropriate noise values
                # Noisy reference pixels for each side of detector
                rd_ref = r * np.mean(self.rd_noise)
                if w[0] > 0: # lower
                    here[:w[0],:] = rd_ref * np.random.standard_normal((w[0],self.naxis1))
                if w[1] > 0: # upper
                    here[-w[1]:,:] = rd_ref * np.random.standard_normal((w[1],self.naxis1))
                if w[2] > 0: # left
                    here[:,:w[2]] = rd_ref * np.random.standard_normal((self.naxis2,w[2]))
                if w[3] > 0: # right
                    here[:,-w[3]:] = rd_ref * np.random.standard_normal((self.naxis2,w[3]))
        
                # Add the noise in to the result
                result[z,:,:] += here
                

        # Add correlated pink noise.
        if self.c_pink is not None:
            # c_pink_map was used to hold the entire correlated pink noise cube
            # c_pink_map was useful for debugging, but eats up a lot of space
            #self.c_pink_map = np.zeros((self.naxis3,self.naxis2,self.naxis1))
            self.message('Adding c_pink noise')
            tt = self.c_pink * self.pink_noise('pink') # tt is a temp. variable
            #self.c_pink_full = tt.copy()
    
            tt = np.reshape(tt, (self.naxis3, self.naxis2+self.nfoh, \
                                 self.xsize+self.nroh))[:,:self.naxis2,:self.xsize]
            for op in np.arange(self.n_out):
                x1 = op * self.xsize
                x2 = x1 + self.xsize

                # By default fast-scan readout direction is [-->,<--,-->,<--]
                # If reverse_scan_direction is True, then [<--,-->,<--,-->]
                # same_scan_direction: all --> or all <--
                if self.same_scan_direction:
                    flip = True if self.reverse_scan_direction else False
                elif np.mod(ch,2)==0:
                    flip = True if self.reverse_scan_direction else False
                else:
                    flip = False if self.reverse_scan_direction else True

                if flip: 
                    #self.c_pink_map[:,:,x1:x2] = tt[:,:,::-1]
                    result[:,:,x1:x2] += tt[:,:,::-1]
                else:
                    #self.c_pink_map[:,:,x1:x2] = tt
                    result[:,:,x1:x2] += tt

            del tt
            #result += self.c_pink_map
            #del self.c_pink_map


        # Add uncorrelated pink noise. Because this pink noise is stationary and
        # different for each output, we don't need to flip it (but why not?)
        if self.u_pink is not None:
            # We want self.u_pink to be an array or list
            if isinstance(self.u_pink, (np.ndarray,list)):
                temp = np.asarray(self.u_pink)
                if temp.size != self.n_out:
                    _log.warning('Number of elements in u_pink not equal to n_out')
                    os.sys.exit()
            else: # Single value as opposed to an array or list
                self.u_pink = np.ones(self.n_out) * self.u_pink
        
            # Only do the rest if any values are not 0
            if self.u_pink.any():
                # u_pink_map was used to hold the entire correlated pink noise cube
                # u_pink_map was useful for debugging, but eats up a lot of space
                #self.u_pink_map = np.zeros((self.naxis3,self.naxis2,self.naxis1))
                self.message('Adding u_pink noise')
                for op in np.arange(self.n_out):
                    x1 = op * self.xsize
                    x2 = x1 + self.xsize
                    tt = self.u_pink[op] * self.pink_noise('pink')
                    tt = np.reshape(tt, (self.naxis3, self.naxis2+self.nfoh, \
                                     self.xsize+self.nroh))[:,:self.naxis2,:self.xsize]

                    if self.same_scan_direction:
                        flip = True if self.reverse_scan_direction else False
                    elif np.mod(ch,2)==0:
                        flip = True if self.reverse_scan_direction else False
                    else:
                        flip = False if self.reverse_scan_direction else True

                    if flip: 
                        #self.u_pink_map[:,:,x1:x2] = tt[:,:,::-1]
                        result[:,:,x1:x2] += tt[:,:,::-1]
                    else:
                        #self.u_pink_map[:,:,x1:x2] = tt
                        result[:,:,x1:x2] += tt

                    del tt
                #result += self.u_pink_map
                #del self.u_pink_map


        # Add ACN
        if self.acn is not None:
            self.message('Adding acn noise')
            for op in np.arange(self.n_out):

                # Generate new pink noise for each even and odd vector.
                # We give these the abstract names 'a' and 'b' so that we
                # can use a previously worked out formula to turn them
                # back into an image section.
                a = self.acn * self.pink_noise('acn')
                b = self.acn * self.pink_noise('acn')

                # Pick out just the real pixels (i.e. ignore the gaps)
                a = a[np.where(self.m_short == 1)]
                b = b[np.where(self.m_short == 1)]

                # Reformat into an image section. This uses the formula
                # mentioned above.
                acn_cube = np.reshape(np.transpose(np.vstack((a,b))),
                                      (self.naxis3,self.naxis2,self.xsize))

                # Add in the ACN. Because pink noise is stationary, we can
                # ignore the readout directions. There is no need to flip
                # acn_cube before adding it in.
                x0 = op * self.xsize
                x1 = x0 + self.xsize
                result[:,:,x0:x1] += acn_cube
                del acn_cube



        # Add PCA-zero. The PCA-zero template is modulated by 1/f.
        if self.pca0_amp is not None:
            self.message('Adding PCA-zero "picture frame" noise')
            gamma = self.pink_noise(mode='pink')
            zoom_factor = self.naxis2 * self.naxis3 / np.size(gamma)
            gamma = zoom(gamma, zoom_factor, order=1, mode='mirror')
            gamma = np.reshape(gamma, (self.naxis3,self.naxis2))
            for z in np.arange(self.naxis3):
                for y in np.arange(self.naxis2):
                    result[z,y,:] += self.pca0_amp*self.pca0[y,:]*gamma[z,y]



        # Add in dark current for each frame
        #k = np.array([[0,0.01,0],[0.01,0.96,0.01],[0,0.01,0]])
        if self.dark_image is not None:
            self.message('Adding dark current')
            gain_temp = self.gain # 2.0 # Temporary gain (e-/ADU)
            dark_frame = self.dark_image * self.tframe * gain_temp # electrons
            # Set reference pixels' dark current equal to 0
            if w[0] > 0: # lower
                dark_frame[:w[0],:] = 0
            if w[1] > 0: # upper
                dark_frame[-w[1]:,:] = 0
            if w[2] > 0: # left
                dark_frame[:,:w[2]] = 0
            if w[3] > 0: # right
                dark_frame[:,-w[3]:] = 0
    
            # For each read frame, create random dark current instance based on Poisson
            dark_temp = np.zeros([self.naxis2,self.naxis1])		
            for z in np.arange(self.naxis3):
                dark_temp += np.random.poisson(dark_frame, size=None)
                #dark_ipc = convolve(dark_temp, k, mode='constant', cval=0.0)
                result[z,:,:] += dark_temp
        

        # If the data cube has only 1 frame, reformat into a 2-dimensional image.
        if self.naxis3 == 1:
            self.message('Reformatting cube into image')
            result = result[0,:,:]


        # Convert to ADU and unsigned int
        if out_ADU:
            # Apply Gain (e/ADU) to convert to ADU (DN)
            if self.gain != 1:
                self.message('Applying Gain')
                result /= self.gain

            # If the data cube has more than one frame, convert to unsigned int
            #if self.naxis3 > 1:
            self.message('Converting to 16-bit unsigned integer')
            # Ensure that there are no negative pixel values. 
            result[result < 0] = 0
            # And that anything higher than 65535 gets tacked to the top end
            result[result >= 2**16] = 2**16 - 1
            result = result.astype('uint16')

        # Create HDU
        # THIS NEEDS TO BE UPDATED WITH BETTER INFO/FORMAT
        hdu = fits.PrimaryHDU(result)
#         hdu.header.append()
#         hdu.header.append(('TFRAME', self.tframe, 'Time in seconds between frames'))
#         hdu.header.append(('INTTIME', self.inttime, 'Total integration time for one MULTIACCUM'))
#         hdu.header.append(('RD_NOISE', self.rd_noise, 'Read noise'))
#         #hdu.header.append(('PEDESTAL', self.pedestal, 'Pedestal drifts'))
#         hdu.header.append(('C_PINK', self.c_pink, 'Correlated pink'))
#         #hdu.header.append(('U_PINK', self.u_pink, 'Uncorrelated pink'))
#         hdu.header.append(('ACN', self.acn, 'Alternating column noise'))
#         hdu.header.append(('PCA0', self.pca0_amp, \
#                            'PCA zero, AKA picture frame'))
        hdu.header['HISTORY'] = 'Created by NGHXRG version ' \
                                + str(self.nghxrg_version)
                        
        # Write the result to a FITS file
        if o_file is not None:
            self.message('Writing FITS file')
            hdu.writeto(o_file, clobber='True')
    
        self.message('Exiting mknoise()')

        return hdu


import numpy as np
import logging
_log = logging.getLogger('pynrc')

# Import libraries
import numpy as np
import pynrc
from pynrc.maths import robust
from pynrc import DetectorOps, setup_logging, conf
from scipy.signal import savgol_filter

class NRC_refs(object):

    """Reference pixel correction object
    
    Object class for reference pixel correction of NIRCam data (single integration).
    Specify the data cube, header, and whether or not the header is in DMS format.

    General usage of functions:
    
    1. Create instance: ``ref = NRC_refs(data, header)``
    2. Determine reference offset values: ``ref.calc_avg_amps()``.
       Stored at ``ref.refs_amps_avg``.
    3. Fix amplifier offsets: ``ref.correct_amp_refs()``.
       Removes offsets that are stored at ``ref.refs_amps_avg``.
    4. Determine average of column references tracking 1/f noise: ``ref.calc_avg_cols()``.
       Reference values offset for a mean value of 0.
       Averages are stored at ``ref.refs_side_avg``.
    5. Optimal smoothing of side reference values: ``ref.calc_col_smooth()``.
       Stores smoothed version at ``ref.refs_side_smth``.
    6. Remove approximation of 1/f noise: ``ref.correct_col_refs()``.


    Parameters
    ----------
    data : ndarray
        Input datacube. Can be two or three dimensions (nz,ny,nx).
    header : obj
        NIRCam Header associated with data.
    DMS : bool
        Is the header in DMS format?
    altcol : bool
        Calculate separate reference values for even/odd columns? 
        Default=True.
    do_all : bool
        Perform the default pixel correction procedures.
    
    """

    def __init__(self, data, header, DMS=True, altcol=True, do_all=False, **kwargs):
    
        # Convert to float if necessary
        if 'float' not in data.dtype.name:
            type_in = data.dtype.name
            data = data.astype(float, copy=False)
            type_out = data.dtype.name
            #print('Converting data from {} to {}'.format(type_in, type_out))

        # Check the number of dimensions are valid.
        ndim = len(data.shape)
        if ndim==2:
            ny,nx = data.shape
            data = data.reshape((1,ny,nx))
            print('Reshaping data to 3 dimensions (nz,ny,nx)')
        elif ndim==3:
            pass 
        else:
            raise ValueError('Input data can only have 2 or 3 dimensions. \
                              Found {} dimensions.'.format(ndim))
                              
        self.data = data
        self.header = header

        self.DMS = DMS
        self.altcol = altcol            

        # Create a detector class
        self._create_detops(**kwargs)

        # Reference info from header
        ref_all = self.detector.ref_info
        self.nref_t = self.header.get('TREFROW', ref_all[1])
        self.nref_b = self.header.get('BREFROW', ref_all[0])
        self.nref_l = self.header.get('LREFCOL', ref_all[2])
        self.nref_r = self.header.get('RREFCOL', ref_all[3])
    
        # Check that reference pixels match up correctly between header and det class
        assert self.nref_t == ref_all[1], 'Number of top reference rows do not match.'
        assert self.nref_b == ref_all[0], 'Number of bottom reference rows do not match.'
        assert self.nref_l == ref_all[2], 'Number of left reference columns do not match.'
        assert self.nref_r == ref_all[3], 'Number of right reference columns do not match.'
    
        # Set amplifier offset values to None initially
        self.refs_amps_avg = None
        # Set column reference values to None initially        
        self.refs_side_avg = None
        
        # Perform all the usual ref pixel corrections with defaults
        if do_all:
            self.calc_avg_amps()
            self.correct_amp_refs()
            self.calc_avg_cols()
            self.calc_col_smooth()
            self.correct_col_refs()
        
    def _create_detops(self, read_mode=None, nint=None, ngroup=None, detector=None, 
        wind_mode=None, xpix=None, ypix=None, x0=None, y0=None, nff=None, **kwargs):
        """
        Create a detector class based on header settings.
        """
    
        from ..detops import create_detops

        header = self.header
        DMS = self.DMS
    
        det = create_detops(header, DMS=DMS, read_mode=read_mode, nint=nint, ngroup=ngroup,
            detector=detector, wind_mode=wind_mode, xpix=xpix, ypix=ypix, x0=x0, y0=y0,
            nff=nff, **kwargs)

        self.detector = det

        # # Detector ID
        # if detector is None:
        #     detector = header.get('SCA_ID')
        #     if detector is None:
        #         detector = header.get('DETECTOR')           
    
        # # Detector size
        # xpix = header['SUBSIZE1'] if DMS else header['NAXIS1'] if xpix is None else xpix
        # ypix = header['SUBSIZE2'] if DMS else header['NAXIS2'] if ypix is None else ypix
    
        # # Subarray position
        # # Headers are 1-indexed, while detector class is 0-indexed
        # if x0 is None:
        #     x1 = header['SUBSTRT1'] if DMS else header['COLCORNR']
        #     x0 = x1 - 1
        # if y0 is None:
        #     y1 = header['SUBSTRT2'] if DMS else header['ROWCORNR']
        #     y0 = y1 - 1
            
        # # Subarray setting: Full, Stripe, or Window
        # if wind_mode is None:
        #     if xpix==ypix==2048:
        #         wind_mode = 'FULL'
        #     else:
        #         # Turn off log warnings
        #         log_prev = conf.logging_level
        #         setup_logging('ERROR', verbose=False)
        #         # Test if STRIPE or WINDOW
        #         det_stripe = DetectorOps(detector, 'STRIPE', xpix, ypix, x0, y0)
        #         det_window = DetectorOps(detector, 'WINDOW', xpix, ypix, x0, y0)
        #         dt_stripe = np.abs(header['TFRAME'] - det_stripe.time_frame)
        #         dt_window = np.abs(header['TFRAME'] - det_window.time_frame)
        #         wind_mode = 'STRIPE' if dt_stripe<dt_window else 'WINDOW'
        #         # Restore previous log levels
        #         setup_logging(log_prev, verbose=False)

        # # Add MultiAccum info
        # hnames = ['READPATT', 'NINTS', 'NGROUPS'] if DMS else ['READOUT',  'NINT',  'NGROUP']

        # read_mode = header[hnames[0]] if read_mode is None else read_mode
        # nint      = header[hnames[1]] if nint      is None else nint
        # ngroup    = header[hnames[2]] if ngroup    is None else ngroup

        # ma_args = {'read_mode':read_mode, 'nint':nint, 'ngroup':ngroup}
                
        # # Create detector class
        # self.detector = DetectorOps(detector, wind_mode, xpix, ypix, x0, y0, **ma_args)

    @property
    def multiaccum(self):
        """A :class:`~pynrc.multiaccum` object"""
        return self.detector.multiaccum
    @property
    def multiaccum_times(self):
        """Exposure timings in dictionary"""
        return self.detector.times_to_dict()

    @property
    def refs_bot(self):
        """Return raw bottom reference values"""
        if self.nref_l>0:
            return self.data[:,:self.nref_b,:]
        else:
            return None
    @property
    def refs_top(self):
        """Return raw top reference values"""
        if self.nref_l>0:
            return self.data[:,-self.nref_t:,:]
        else:
            return None
    @property
    def refs_right(self):
        """Return raw right reference values"""
        if self.nref_l>0:
            return self.data[:,:,-self.nref_r:]
        else:
            return None
    @property
    def refs_left(self):
        """Return raw left reference values"""
        if self.nref_l>0:
            return self.data[:,:,:self.nref_l]
        else:
            return None

    @property
    def mask_ref(self):
        """Reference pixel mask for det coordinates"""
        return self.detector.mask_ref
    @property
    def mask_act(self):
        """Active pixel mask for det coordinates"""
        return self.detector.mask_act
    @property
    def mask_channels(self):
        """Channel masks for det coordinates"""
        return self.detector.mask_channels

    def calc_avg_amps(self, top_ref=True, bot_ref=True):
        """Calculate amplifier averages
        
        Save the average reference value for each amplifier in each frame.
        Each array has a size of (namp, ngroup). Average values are 
        saved at ``self.refs_amps_avg``. 
        
        Parameters
        ----------
        top_ref : bool
            Include top reference rows when correcting channel offsets.
        bot_ref : bool
            Include bottom reference rows when correcting channel offsets.
        """
        nchans = self.detector.nout
        #chsize = self.detector.chsize
        data_shape = self.data.shape

        if self.nref_t==0: top_ref = False
        if self.nref_b==0: bot_ref = False

        if top_ref and bot_ref:
            refs_all = np.hstack((self.refs_bot, self.refs_top))
        elif bot_ref and (not top_ref):
            refs_all = self.refs_bot
        elif top_ref and (not bot_ref):
            refs_all = self.refs_top
        else:
            print("No top or bottom reference pixels to calculate offset values.")
            #self.refs_amps_avg = None
            return
            
        self.refs_amps_avg = calc_avg_amps(refs_all, data_shape, nchans, self.altcol)
        self.supermean = robust.mean(refs_all)

    def correct_amp_refs(self, supermean=False):
        """Correct amplifier offsets
        
        Use values in ``self.refs_amps_avg`` to correct amplifier offsets.

        Parameters
        ----------
        supermean : bool
            Add back the overall mean of the reference pixels.
        """
    
        # Check to make sure refs_amps_avg is valid
        if (self.refs_amps_avg is None):
            _log.warning('self.refs_amps_avg is set to None. No offsets applied.')
            return
            #raise ValueError('self.refs_amps_avg is set to None')

        # Supermean
        # the average of the average is the DC level of the output channel
        smean = self.supermean if supermean else 0.0
    
        nchans = self.detector.nout
        chsize = self.detector.chsize
        nz, ny, nx = self.data.shape
        for ch in range(nchans):
            # Channel indices
            ich1 = int(ch*chsize)
            ich2 = int(ich1 + chsize)
        
            # In-place subtraction of channel averages
            if self.altcol:
                for i in range(nz):
                    self.data[i,:,ich1:ich2-1:2] -= self.refs_amps_avg[0][ch,i]
                    self.data[i,:,ich1+1:ich2:2] -= self.refs_amps_avg[1][ch,i]
            else:
                for i in range(nz):
                    self.data[i,:,ich1:ich2] -= self.refs_amps_avg[ch,i]

        # Add back supermean
        if supermean: self.data += smean
        
    
    def calc_avg_cols(self, left_ref=True, right_ref=True, avg_type='frame', **kwargs):
        """Calculate average of column references
        
        Create a copy of the left and right reference pixels, removing the 
        average value of the reference pixels on an int, frame, or pixel basis. 
        Do this after correcting the amplifier offsets with ``correct_amp_refs()``.
        Averages are stored in ``self.refs_side_avg``.
        
        Parameters
        ----------
        left_ref : bool
            Include left reference cols when correcting 1/f noise.
        right_ref : bool
            Include right reference cols when correcting 1/f noise.
        avg_type : str
            Type of ref col averaging to perform. Allowed values are
            'pixel', 'frame', or 'int'.
        mean_func : func
            Function to use to calculate averages of reference columns
        """
        
        if self.nref_l==0: left_ref = False
        if self.nref_r==0: right_ref = False
        
        if (not left_ref) and (not right_ref):
            print("No left or right reference pixels to calculate 1/f noise.")
            self.refs_side_avg = None
            return

        rl = self.refs_left  if left_ref  else None
        rr = self.refs_right if right_ref else None
        self.refs_side_avg = calc_avg_cols(rl, rr, avg_type, **kwargs)
                    

    def calc_col_smooth(self, perint=False, edge_wrap=False, savgol=False, **kwargs):
        """Optimal smoothing of side reference pixels
        
        Geneated smoothed version of column reference values.
        Uses :func:`calc_avg_cols` to determine approx 1/f noise in data
        and store in ``self.refs_side_smth``.
        
        Parameters
        ----------
        perint : bool
            Smooth side reference pixel per int, otherwise per frame.
        edge_wrap : bool
            Add a partial frames to the beginning and end of each averaged
            time series pixels in order to get rid of edge effects.          
        """
        
        refvals = self.refs_side_avg

        # Check to make sure refs_amps_avg1 and refs_amps_avg2 are valid
        if refvals is None:
            _log.warning('self.refs_side_avg is set to None. No smoothing applied.')
            return
            #raise ValueError('self.refs_side_avg set to None')
        
        # Time to go through an entire row.
        # The delta time does't seem to make any difference in the final data product
        # Just for vizualization purposes...
        xticks = self.detector.chsize + self.detector._line_overhead
        delt = xticks / self.detector._pixel_rate
        
        # Save smoothed values
        self.refs_side_smth = calc_col_smooth(refvals, self.data.shape, \
                                              perint=perint, edge_wrap=edge_wrap, 
                                              delt=delt, savgol=savgol, **kwargs)
    
    def correct_col_refs(self):
        """Remove 1/f noise from data
        
        Correct 1/f noise using the approximation stored in 
        ``self.refs_side_smth``.
        """
        
        # Final correction
        #for i,im in enumerate(cube): im -= refvals_smoothed[i].reshape([ny,1])
        nz, ny, nx = self.data.shape
        self.data -= self.refs_side_smth.reshape([nz,ny,1])
        


def reffix_hxrg(cube, nchans=4, in_place=True, fixcol=False, **kwargs):
    """Reference pixel correction function
    
    This function performs a reference pixel correction
    on HAWAII-[1,2,4]RG detector data read out using N outputs.
    Top and bottom reference pixels are used first to remove 
    channel offsets.

    Parameters
    ----------
    cube : ndarray
        Input datacube. Can be two or three dimensions (nz,ny,nx).
    in_place : bool
        Perform calculations in place. Input array is overwritten.
    nchans : int
        Number of output amplifier channels in the detector. Default=4.
    fixcol : bool
        Perform reference column corrections?
        
    Keyword Args
    ------------
    altcol : bool
        Calculate separate reference values for even/odd columns.
    supermean : bool
        Add back the overall mean of the reference pixels.
    top_ref : bool
        Include top reference rows when correcting channel offsets.
    bot_ref : bool
        Include bottom reference rows when correcting channel offsets.
    ntop : int
        Specify the number of top reference rows.
    nbot : int
        Specify the number of bottom reference rows.
    left_ref : bool
        Include left reference cols when correcting 1/f noise.
    right_ref : bool
        Include right reference cols when correcting 1/f noise.
    nleft : int
        Specify the number of left reference columns.
    nright : int
        Specify the number of right reference columns.
    perint : bool
        Smooth side reference pixel per integration, otherwise do frame-by-frame.
    avg_type :str
        Type of side column averaging to perform to determine ref pixel drift. 
        Allowed values are 'pixel', 'frame', or 'int':

            * 'int'   : Subtract the avg value of all side ref pixels in ramp.
            * 'frame' : For each frame, get avg of side ref pixels and subtract framewise.
            * 'pixel' : For each ref pixel, subtract its avg value from all frames.

    savgol : bool
        Using Savitsky-Golay filter method rather than FFT.
    winsize : int
        Size of the window filter.
    order : int
        Order of the polynomial used to fit the samples.
    """

    # Check the number of dimensions are valid.
    ndim = len(cube.shape)
    if not (ndim==2 or ndim==3):
        raise ValueError('Input data can only have 2 or 3 dimensions. \
                          Found {} dimensions.'.format(ndim))

    # Convert to float
    if 'float' not in cube.dtype.name:
        copy = (not in_place)
        cube = cube.astype(float, copy=copy)

    if not in_place:
        cube = np.copy(cube)

    # Remove channel offsets
    cube = reffix_amps(cube, nchans=nchans, in_place=True, **kwargs)

    # Fix 1/f noise using vertical reference pixels
    if fixcol:
        cube = ref_filter(cube, nchans=nchans, in_place=True, **kwargs)

    return cube
    
    
def reffix_amps(cube, nchans=4, in_place=True, altcol=True, supermean=False,
    top_ref=True, bot_ref=True, ntop=4, nbot=4, **kwargs):
    """Correct amplifier offsets
    
    Matches all amplifier outputs of the detector to a common level.

    This routine subtracts the average of the top and bottom reference rows
    for each amplifier and frame individually.

    By default, reference pixel corrections are performed in place since it's
    faster and consumes less memory.

    Parameters
    ----------
    cube : ndarray
        Input datacube. Can be two or three dimensions (nz,ny,nx).
    nchans : int
        Number of output amplifier channels in the detector. Default=4.
    altcol : bool
        Calculate separate reference values for even/odd columns.
    supermean : bool
        Add back the overall mean of the reference pixels.
    in_place : bool
        Perform calculations in place. Input array is overwritten.
    top_ref : bool
        Include top reference rows when correcting channel offsets.
    bot_ref : bool
        Include bottom reference rows when correcting channel offsets.
    ntop : int
        Specify the number of top reference rows.
    nbot : int
        Specify the number of bottom reference rows.
    """

    if not in_place:
        cube = np.copy(cube)

    # Check the number of dimensions are valid.
    ndim = len(cube.shape)
    if ndim==2:
        ny,nx = cube.shape
        nz = 1
        cube = cube.reshape((nz,ny,nx))
    elif ndim==3:
        nz, ny, nx = cube.shape
    else:
        raise ValueError('Input data can only have 2 or 3 dimensions. \
                          Found {} dimensions.'.format(ndim))        

    chsize = int(nx / nchans)

    # Number of reference rows to use
    # Set nt or nb equal to 0 if we don't want to use either
    nt = ntop if top_ref else 0
    nb = nbot if bot_ref else 0

    if (nt+nb)==0: 
        print("No reference pixels available for use. Returning...")
        return

    # Slice out reference pixels
    refs_bot = cube[:,:nb,:]
    refs_top = cube[:,-nt:,:]
    if nt==0:
        refs_all = refs_bot
    elif nb==0:
        refs_all = refs_top
    else:
        refs_all = np.hstack((refs_bot, refs_top))
    
    assert refs_all.shape[1] == (nb+nt)

    # Supermean
    # the average of the average is the DC level of the output channel
    smean = robust.mean(refs_all) if supermean else 0.0
    
    # Calculate avg reference values for each frame and channel
    refs_amps_avg = calc_avg_amps(refs_all, cube.shape, nchans=nchans, altcol=altcol)
        
    for ch in range(nchans):
        # Channel indices
        ich1 = ch*chsize
        ich2 = ich1 + chsize

        # In-place subtraction of channel medians
        if altcol:
            for i in range(nz):
                cube[i,:,ich1:ich2-1:2] -= refs_amps_avg[0][ch,i]
                cube[i,:,ich1+1:ich2:2] -= refs_amps_avg[1][ch,i]
        else:
            for i in range(nz):
                cube[i,:,ich1:ich2] -= refs_amps_avg[ch,i]

    # Add back supermean
    if supermean: cube += smean

    cube = cube.squeeze()
    return cube


def ref_filter(cube, nchans=4, in_place=True, avg_type='frame', perint=False, 
    edge_wrap=False, left_ref=True, right_ref=True, nleft=4, nright=4, **kwargs):
    """Optimal Smoothing
    
    Performs an optimal filtering of the vertical reference pixel to 
    reduce 1/f noise (horizontal stripes).

    Adapted from M. Robberto IDL code:
    http://www.stsci.edu/~robberto/Main/Software/IDL4pipeline/

    Parameters
    ----------
    cube : ndarray
        Input datacube. Can be two or three dimensions (nz,ny,nx).
    nchans : int
        Number of output amplifier channels in the detector. Default=4.
    in_place : bool
        Perform calculations in place. Input array is overwritten.    
    perint : bool
        Smooth side reference pixel per integration, 
        otherwise do frame-by-frame.
    avg_type : str
        Type of ref col averaging to perform. Allowed values are
        'pixel', 'frame', or 'int'.
    left_ref : bool
        Include left reference cols when correcting 1/f noise.
    right_ref : bool
        Include right reference cols when correcting 1/f noise.
    nleft : int
        Specify the number of left reference columns.
    nright : int
        Specify the number of right reference columns.

    Keyword Arguments
    =================
    savgol : bool
        Using Savitsky-Golay filter method rather than FFT.
    winsize : int
        Size of the window filter.
    order : int
        Order of the polynomial used to fit the samples.
    mean_func : func
        Function to use to calculate averages of reference columns.
    """               
           
    if not in_place:
        cube = np.copy(cube)

    # Check the number of dimensions are valid.
    ndim = len(cube.shape)
    if ndim==2:
        ny,nx = cube.shape
        nz = 1
        cube = cube.reshape((nz,ny,nx))
    elif ndim==3:
        nz, ny, nx = cube.shape
    else:
        raise ValueError('Input data can only have 2 or 3 dimensions. \
                          Found {} dimensions.'.format(ndim))        

    # Number of reference rows to use
    # Set nt or nb equal to 0 if we don't want to use either
    nl = nleft  if left_ref  else 0
    nr = nright if right_ref else 0

    assert nl>=0, 'Number of left reference pixels must not be negative.'
    assert nr>=0, 'Number of right reference pixels must not be negative.'

    if (nl+nr)==0: 
        print("No reference pixels available for use. Returning...")
        return
        
    # Slice out reference pixel columns
    refs_left  = cube[:,:,:nl]  if nl>0 else None
    refs_right = cube[:,:,-nr:] if nr>0 else None
    refvals = calc_avg_cols(refs_left, refs_right, avg_type, **kwargs)

    # The delta time does't seem to make any difference in the final data product
    # Just for vizualization purposes...
    delt = 10E-6 * (nx/nchans + 12.)
    refvals_smoothed = calc_col_smooth(refvals, cube.shape, perint=perint, 
        edge_wrap=edge_wrap, delt=delt, **kwargs)
    
    # Final correction
    #for i,im in enumerate(cube): im -= refvals_smoothed[i].reshape([ny,1])
    cube -= refvals_smoothed.reshape([nz,ny,1])

    cube = cube.squeeze()
    return cube

    

def calc_avg_amps(refs_all, data_shape, nchans=4, altcol=True):
    """Calculate amplifier averages
    
    Save the average reference value for each amplifier in each frame.
    Assume by default that alternating columns are offset from each other,
    so we save two arrays: self.refs_amps_avg1 and self.refs_amps_avg2. 
    Each array has a size of (namp, ngroup).

    Parameters
    ----------
    refs_all : ndarray
        The top and/or bottom references pixels order 
        in a shape (nz, nref_rows, nx)
    data_shape : tuple
        Shape of the data array: (nz, ny, nx).
    nchans : int
        Number of amplifier output channels.
    altcol : bool
        Calculate separate reference values for even/odd columns? 
        Default=True.
    """
        
    nz, ny, nx = data_shape
    chsize = int(nx / nchans)
    
    if altcol:
        refs_amps_avg1 = []
        refs_amps_avg2 = []
        for ch in range(nchans):
            # Channel indices
            ich1 = ch*chsize
            ich2 = ich1 + chsize

            # Slice out alternating columns
            refs_ch1 = refs_all[:,:,ich1:ich2-1:2].reshape((nz,-1))
            refs_ch2 = refs_all[:,:,ich1+1:ich2:2].reshape((nz,-1))

            # Take the resistant mean
            chavg1 = robust.mean(refs_ch1,axis=1)
            chavg2 = robust.mean(refs_ch2,axis=1)
    
            refs_amps_avg1.append(chavg1)
            refs_amps_avg2.append(chavg2)

        return (np.array(refs_amps_avg1), np.array(refs_amps_avg2))
    else:
        refs_amps_avg = []
        for ch in range(nchans):
            # Channel indices
            ich1 = ch*chsize
            ich2 = ich1 + chsize

            # Slice out alternating columns
            refs_ch = refs_all[:,:,ich1:ich2].reshape((nz,-1))

            # Take the resistant mean and reshape for broadcasting
            chavg = robust.mean(refs_ch,axis=1).reshape(-1,1,1)
            refs_amps_avg.append(chavg)
            
        return np.array(refs_amps_avg)
        
        
def calc_avg_cols(refs_left=None, refs_right=None, avg_type='frame',
    mean_func=np.median, **kwargs):
    """Calculate average of column references
    
    Determine the average values for the column references, which
    is subsequently used to estimate the 1/f noise contribution.

    Parameters
    ----------
    refs_left : ndarray
        Left reference columns.
    refs_right : ndarray
        Right reference columns.
    avg_type : str
        Type of ref column averaging to perform to determine ref pixel variation. 
        Allowed values are 'pixel', 'frame', or 'int'.
        'pixel' : For each ref pixel, subtract its avg value from all frames.
        'frame' : For each frame, get avg ref pixel values and subtract framewise.
        'int'   : Calculate avg of all ref pixels within the ramp and subtract.
    mean_func : func
        Function to use to calculate averages of reference columns
    """
    
    # Which function to use for calculating averages?
    # mean_func = robust.mean
    # mean_func = np.median

    # In this context, nl and nr are either 0 (False) or 1 (True)
    nl = 0 if refs_left is None else 1
    nr = 0 if refs_right is None else 1

    # Left and right reference pixels
    # Make a copy so as to not modify the original data?
    if nl>0: refs_left  = np.copy(refs_left)
    if nr>0: refs_right = np.copy(refs_right)

    # Set the average of left and right reference pixels to zero
    # By default, pixel averaging is best for large groups
    if avg_type is None:
        avg_type = 'frame'

    if refs_left is not None:
        nz, ny, nchan = refs_left.shape
    else:
        nz, ny, nchan = refs_right.shape

    # If there is only 1 frame, then we have to do "per frame" averaging.
    # Set to "per int", which produces the same result as "per frame" for nz=1.
    if nz==1:
        avg_type = 'int'

    # Remove average ref pixel values
    # Average over entire integration
    if 'int' in avg_type:
        if nl>0: refs_left  -= mean_func(refs_left)
        if nr>0: refs_right -= mean_func(refs_right)
    # Average over each frame
    elif 'frame' in avg_type:
        if nl>0: refs_left_mean  = mean_func(refs_left.reshape((nz,-1)), axis=1)
        if nr>0: refs_right_mean = mean_func(refs_right.reshape((nz,-1)), axis=1)
        # Subtract estimate of each ref pixel "intrinsic" value
        for i in range(nz):
            if nl>0: refs_left[i]  -= refs_left_mean[i]
            if nr>0: refs_right[i] -= refs_right_mean[i]
    # Take the average of each reference pixel 
    elif 'pix' in avg_type:
        if nl>0: refs_left_mean  = mean_func(refs_left, axis=0)
        if nr>0: refs_right_mean = mean_func(refs_right, axis=0)
        # Subtract estimate of each ref pixel "intrinsic" value
        for i in range(nz):
            if nl>0: refs_left[i]  -= refs_left_mean
            if nr>0: refs_right[i] -= refs_right_mean

    if nl==0:
        refs_side_avg = refs_right.mean(axis=2)
    elif nr==0:
        refs_side_avg = refs_left.mean(axis=2)    
    else:
        refs_side_avg = (refs_right.mean(axis=2) + refs_left.mean(axis=2)) / 2
        
    return refs_side_avg



def calc_col_smooth(refvals, data_shape, perint=False, edge_wrap=False, 
    delt=5.24E-4, savgol=False, winsize=31, order=3, **kwargs):
    """Perform optimal smoothing of side ref pix
    
    Generates smoothed version of column reference values.
    Smooths values from calc_avg_cols() via FFT.

    Parameters
    ----------
    refvals : ndarray
        Averaged column reference pixels
    data_shape : tuple
        Shape of original data (nz,ny,nx)
        
    Keyword Arguments
    =================
    perint : bool
        Smooth side reference pixel per int, otherwise per frame.
    edge_wrap : bool
        Add a partial frames to the beginning and end of each averaged
        time series pixels in order to get rid of edge effects.
    delt : float
        Time between reference pixel samples. 
    savgol : bool
        Using Savitsky-Golay filter method rather than FFT.
    winsize : int
        Size of the window filter.
    order : int
        Order of the polynomial used to fit the samples.
    """
    
    nz,ny,nx = data_shape
    
    # May want to revisit the do-all-at-once or break-up decision
    # This may depend on preamp reset per frame or per integration
    # For now, we'll do it frame-by-frame by default (perint=False)
    if perint: # per integration
        if edge_wrap: # Wrap around to avoid edge effects
            refvals2 = np.vstack((refvals[0][::-1], refvals, refvals[-1][::-1]))
            if savgol: # SavGol filter
                refvals_smoothed2 = savgol_filter(refvals2.ravel(), winsize, order, delta=1)
            else: # Or "optimal" smoothing algorithm
                refvals_smoothed2 = smooth_fft(refvals2, delt)
            refvals_smoothed = refvals_smoothed2[ny:-ny].reshape(refvals.shape)
        else:
            if savgol:
                refvals_smoothed = savgol_filter(refvals.ravel(), winsize, order, delta=1)
            else:
                refvals_smoothed = smooth_fft(refvals, delt)
            refvals_smoothed = refvals_smoothed.reshape(refvals.shape)


    else:
        refvals_smoothed = []
        if edge_wrap: # Wrap around to avoid edge effects
            for ref in refvals:
                ref2 = np.concatenate((ref[:ny//2][::-1], ref, ref[ny//2:][::-1]))
                if savgol:
                    ref_smth = savgol_filter(ref2, winsize, order, delta=1)
                else:
                    ref_smth = smooth_fft(ref2, delt)
                refvals_smoothed.append(ref_smth[ny//2:ny//2+ny])
            refvals_smoothed = np.array(refvals_smoothed)
        else:
            for ref in refvals:
                if savgol:
                    ref_smth = savgol_filter(ref, winsize, order, delta=1)
                else: 
                    ref_smth = smooth_fft(ref, delt)
                refvals_smoothed.append(ref_smth)
            refvals_smoothed = np.array(refvals_smoothed)
    
    return refvals_smoothed


    
def smooth_fft(data, delt, first_deriv=False, second_deriv=False):
    """Optimal smoothing algorithm
    
    Smoothing algorithm to perform optimal filtering of the 
    vertical reference pixel to reduce 1/f noise (horizontal stripes),
    based on the Kosarev & Pantos algorithm. This assumes that the
    data to be filtered/smoothed has been sampled evenly.

    If first_deriv is set, then returns two results
    if second_deriv is set, then returns three results.

    Adapted from M. Robberto IDL code:
    http://www.stsci.edu/~robberto/Main/Software/IDL4pipeline/

    Parameters
    ----------
    data : ndarray
        Signal to be filtered.
    delt : float
        Delta time between samples.
    first_deriv : bool
        Return the first derivative.    
    second_deriv : bool
        Return the second derivative (along with first).

    """

    Dat = data.flatten()
    N = Dat.size
    Pi2 = 2*np.pi
    OMEGA = Pi2 / (N*delt)
    X = np.arange(N) * delt

    ##------------------------------------------------
    ## Center and Baselinefit of the data
    ##------------------------------------------------
    Dat_m = Dat - np.mean(Dat)
    SLOPE = (Dat_m[-1] - Dat_m[0]) / (N-2)
    Dat_b = Dat_m - Dat_m[0] - SLOPE * X / delt

    ##------------------------------------------------
    ## Compute fft- / power- spectrum
    ##------------------------------------------------
    Dat_F = np.fft.rfft(Dat_b) #/ N
    Dat_P = np.abs(Dat_F)**2

    # Frequency for abscissa axis
    # F = [0.0, 1.0/(N*delt), ... , 1.0/(2.0*delt)]:
    #F = np.arange(N/2+1) / (N*delt)
    #F = np.fft.fftfreq(Dat_F.size, delt)

    ##------------------------------------------------
    ## Noise spectrum from 'half' to 'full'
    ## Mind: half means N/4, full means N/2
    ##------------------------------------------------
    i1 = int((N-1) / 4)
    i2 = int((N-1) / 2) + 1
    Sigma = np.sum(Dat_P[i1:i2])
    Noise = Sigma / ((N-1)/2 - (N-1)/4)

    ##------------------------------------------------
    ## Get Filtercoeff. according to Kosarev/Pantos
    ## Find the J0, start search at i=1 (i=0 is the mean)
    ##------------------------------------------------
    J0 = 2
    for i in np.arange(1, int(N/4)+1):
        sig0, sig1, sig2, sig3 = Dat_P[i:i+4]
        if (sig0<Noise) and ((sig1<Noise) or (sig2<Noise) or (sig3<Noise)):
            J0 = i
            break

    ##------------------------------------------------
    ## Compute straight line extrapolation to log(Dat_P)
    ##------------------------------------------------
    ii = np.arange(1,J0+1)
    logvals = np.log(Dat_P[1:J0+1])
    XY = np.sum(ii * logvals)
    XX = np.sum(ii**2)
    S  = np.sum(logvals)
    # Find parameters A1, B1
    XM = (2. + J0) / 2
    YM = S / J0
    A1 = (XY - J0*XM*YM) / (XX - J0*XM*XM)
    B1 = YM - A1 * XM

    # Compute J1, the frequency for which straight
    # line extrapolation drops 20dB below noise
    J1 = int(np.ceil((np.log(0.01*Noise) - B1) / A1 ))
    if J1<J0:
        J1 = J0+1


    ##------------------------------------------------
    ## Compute the Kosarev-Pantos filter windows
    ## Frequency-ranges: 0 -- J0 | J0+1 -- J1 | J1+1 -- N2
    ##------------------------------------------------
    nvals = int((N-1)/2 + 1)
    LOPT = np.zeros_like(Dat_P)
    LOPT[0:J0+1] = Dat_P[0:J0+1] / (Dat_P[0:J0+1] + Noise)
    i_arr = np.arange(J1-J0) + J0+1
    LOPT[J0+1:J1+1] = np.exp(A1*i_arr+B1) / (np.exp(A1*i_arr+B1) + Noise)

    ##--------------------------------------------------------------------
    ## De-noise the Spectrum with the filter
    ## Calculate the first and second derivative (i.e. multiply by iW)
    ##--------------------------------------------------------------------

    # first loop gives smoothed data
    # second loop produces first derivative
    # third loop produces second derivative
    if second_deriv:
        ndiff = 3
    elif first_deriv:
        ndiff = 2
    else:
        ndiff = 1

    for diff in range(ndiff):
        #Fltr_Spectrum = np.zeros(N,dtype='complex')
        Fltr_Spectrum = np.zeros_like(Dat_P,dtype='complex')
        # make the filter complex
        i1 = 1; n2 = int((N-1)/2)+1; i2 = i1+n2 
        FltrCoef = LOPT[i1:].astype('complex')
        # differentitation in frequency domain
        iW = ((np.arange(n2)+i1)*OMEGA*1j)**diff
        # multiply spectrum with filter coefficient
        Fltr_Spectrum[i1:] = Dat_F[i1:] * FltrCoef * iW

        # Fltr_Spectrum[0] values
        # The derivatives of Fltr_Spectrum[0] are 0
        # Mean if diff = 0
        Fltr_Spectrum[0] = 0 if diff>0 else Dat_F[0]
    
        # Inverse fourier transform back in time domain
        Dat_T = np.fft.irfft(Fltr_Spectrum)
        #Dat_T[-1] = np.real(Dat_T[0]) + 1j*np.imag(Dat_T[-1])

        # This ist the smoothed time series (baseline added)
        if diff==0:
            Smoothed_Data = np.real(Dat_T) + Dat[0] + SLOPE * X / delt
        elif diff==1:
            First_Diff = np.real(Dat_T) + SLOPE / delt
        elif diff==2:
            Secnd_Diff = np.real(Dat_T)
    
    if second_deriv:
        return Smoothed_Data, First_Diff, Secnd_Diff
    elif first_deriv:
        return Smoothed_Data, First_Diff
    else:
        return Smoothed_Data
    

def chrem_med(imarr, nchans=4, yind=None, bpmask=None, in_place=True,
              mean_func=np.median):
    """ Subtract Amplifier Channel Offsets
    
    Sometimes amplifiers have offsets relative to each other 
    due to imperfect tracking of reference pixels. This function
    determines the average offset from zero of each channel
    and subtracts the mean/median from the entire channel.
    
    Parameters
    ----------
    imarr : ndarray
        Array of image (or single image).
    nchans : int
        Number of amplifier readout channels.
    yind : array-like
        Two element array to select a y-range for calculating
        the channel offset.
    bpmask : bool array
        Bad pixel mask (1 for bad, 0 for good). Can either
        be a single image or image cube of same size as `imarr`.
    in_place : bool
        Correct in-place? If False, returns a copy of the array
        with channels offset.
    mean_func : func
        Function to use for performing the mean calculation.
    
    """
    
    sh_orig = imarr.shape
    if len(sh_orig)==2:
        nz = 1
        ny, nx = sh_orig
        imarr = imarr.reshape([nz,ny,nx])
    else:
        nz, ny, nx = sh_orig    
        
    chsize = int(nx / nchans)
    
    # Make copy of array?
    arr_out = imarr if in_place else imarr.copy()
    
    # Define y index start/stop locations
    yind = np.array([0,ny]) if yind is None else yind
    
    bpmask = np.zeros([ny,nx]) if bpmask is None else bpmask
    bpmask = bpmask.squeeze()
    
    # Cropped array
    for ch in np.arange(nchans):
        # Get channel x-indices
        x1 = int(ch*chsize)
        x2 = int(x1 + chsize)

        # Select the channel and y-range
        imch = arr_out[:, yind[0]:yind[1], x1:x2]
        # imch_ind = imch.reshape([imch.shape[0],-1])
        
        # Take median of all pixels in channel for each image
        if len(bpmask.shape)==2:
            bpmask_ch = bpmask[yind[0]:yind[1], x1:x2]
            igood = (bpmask_ch == 0)
            chmed = mean_func(imch[:,igood], axis=1)

            # Subtract median channel from each image
            arr_out[:,:,x1:x2] -= chmed.reshape([-1,1,1])
        else:
            for jj, im in enumerate(imch): 
                bpmask_ch = bpmask[jj, yind[0]:yind[1], x1:x2]
                igood = (bpmask_ch == 0)
                arr_out[jj,:,x1:x2] -= mean_func(im[igood])
                    
    return arr_out.reshape(sh_orig)


def channel_averaging(im, nchans=4, same_scan_direction=False, off_chans=True, 
    mn_func=np.nanmedian, **kwargs):
    """Estimate common 1/f noise in image

    For a given image, average the channels together to find 
    the common pattern noise present within the channels.
    Returns an array the same size as the input image.

    Parameters
    ==========
    im : ndarray
        Input image

    Keyword Args
    ============
    nchans : int
        Number of output channels
    same_scan_direction : bool
        Are all the output channels read in the same direction?
        By default fast-scan readout direction is ``[-->,<--,-->,<--]``
        If ``same_scan_direction``, then all ``-->``
    off_chans : bool
        Calculate indepenent values for each channel using the off channels.
    mn_func : function
        What function should we use to calculate the average.
        Default `np.nanmedian`

    """

    ny, nx = im.shape
    chsize = int(nx / nchans)

    # Reshape to [ny,chsize,nchans]
    im = im.reshape(ny,nchans,chsize).transpose([0,2,1])

    # Flip channels if they're reversed
    if same_scan_direction==False:
        # Make sure we don't modify the input array
        im = im.copy()
        for ch in range(nchans):
            if np.mod(ch,2)==1: 
                im[:,:,ch] = im[:,::-1,ch]
                
    if off_chans == False:
        im = im.reshape([-1,nchans])
        ch_mn = mn_func(im, axis=1).reshape([ny,chsize])
        im = im.reshape([ny,chsize,nchans])

    arr_list = []
    ind_chans = np.arange(nchans)
    for ch in range(nchans):

        # Take median of other channels
        if off_chans:
            ind_off = np.where(ind_chans != ch)[0]
            im_off_chans = im[:,:,ind_off].reshape([-1,nchans-1])
            ch_mn = mn_func(im_off_chans, axis=1).reshape([ny,chsize])
        
        # Consecutive outputs reversed?
        if (np.mod(ch,2) == 0) or (same_scan_direction == True): 
            arr_list.append(ch_mn)
        else: 
            arr_list.append(ch_mn[:,::-1])
    # im = im.reshape([ny,chsize,nchans]).transpose([0,2,1]).reshape([ny,nx])

    # Flip channels back to original position
    if same_scan_direction==False:
        for ch in range(nchans):
            if np.mod(ch,2)==1: 
                im[:,:,ch] = im[:,::-1,ch]
            
    return np.concatenate(arr_list, axis=1)


def channel_smooth_fft(im_arr, winsize=64):
    """Channel smoothing using smooth_fft

    Function for generating a map of the 1/f noise within a series of input images.
    The input images should show some clear noise structure for this to be useful.
    Uses M. Robberto smoothing algorithm.

    One might prefer the `channel_smooth_savgol` or `channel_smooth_butter`
    functions due to their quickness.

    Parameters
    ==========
    im_arr : ndarray
        Input array of images
    winsize : int
        Window size chunks to break up 
    """

    sh = im_arr.shape
    if len(sh)==2:
        nz = 1
        ny, chsize = sh
    else:
        nz, ny, chsize = sh

    # Check that winsize is even
    winsize = winsize+1 if winsize % 2 == 1 else winsize

    # Reshape in case of nz=1
    im_arr = im_arr.reshape([nz, -1])

    res_arr = []
    excess = winsize #int(winsize / 2)
    for im in im_arr:
        nwin = int(im.size / winsize) + 1
        # Add some extra values to beginning and end to remove edge effects
        im2 = np.concatenate((im[:excess][::-1], im, im[-excess:][::-1]))

        res = []
        for i in range(nwin):
            i1 = 0 if i==0 else winsize*i
            i2 = i1 + winsize + 2*excess
            vals = im2[i1:i2]

            # If smooth_fft fails, then just take median
            # Failing generally means the distribution is consistent with white noise
            try:
                vals_smooth = smooth_fft(vals, 10e-6)
                # Trim edges
                vals_smooth = vals_smooth[excess:excess+winsize]
            except:
                vals_smooth = np.zeros(winsize) + np.nanmedian(vals)
            res.append(vals_smooth)
        res = np.array(res).ravel()[0:im.size]
        res_arr.append(res.reshape([ny,-1]))

    return np.array(res_arr).squeeze()

def mask_helper():
    """Helper to handle indices and logical indices of a mask

    Output: index, a function, with signature indices = index(logical_indices),
    to convert logical indices of a mask to 'equivalent' indices

    Example:
        >>> # linear interpolation of NaNs
        >>> mask = np.isnan(y)
        >>> x = mask_helper(y)
        >>> y[mask]= np.interp(x(mask), x(~mask), y[~mask])
    """

    return lambda z: np.nonzero(z)[0]

def channel_smooth_savgol(im_arr, winsize=31, order=3, per_line=False, 
    mask=None, **kwargs):
    """Channel smoothing using savgol filter

    Parameters
    ==========
    im_arr : ndarray
        Input array of images (intended to be a cube of output channels).
        Shape should either be (ny, chsize) to smooth a single channel or
        (nchan, ny, chsize) for  multiple channels.
        Each image is operated on separately. If only two dimensions,
        then only a single input image is assumed. NaN's will be
        interpolated over.

    Keyword Args
    ============
    winsize : int
        Size of the window filter. Should be an odd number.
    order : int
        Order of the polynomial used to fit the samples.
    per_line : bool
        Smooth each channel line separately with the hopes of avoiding
        edge discontinuities.
    mask : bool image or None
        An image mask of pixels to ignore. Should be same size as im_arr.
        This can be used to mask pixels that the filter should ignore, 
        such as stellar sources or pixel outliers. A value of True indicates
        that pixel should be ignored.
    mode : str
        Must be 'mirror', 'constant', 'nearest', 'wrap' or 'interp'.  This
        determines the type of extension to use for the padded signal to
        which the filter is applied.  When `mode` is 'constant', the padding
        value is given by `cval`. 
        When the 'interp' mode is selected (the default), no extension
        is used.  Instead, a degree `polyorder` polynomial is fit to the
        last `window_length` values of the edges, and this polynomial is
        used to evaluate the last `window_length // 2` output values.
    cval : float
        Value to fill past the edges of the input if `mode` is 'constant'.
        Default is 0.0.
    """

    sh = im_arr.shape
    if len(sh)==2:
        nz = 1
        ny, chsize = sh
    else:
        nz, ny, chsize = sh

    # Check that winsize is odd
    winsize = winsize-1 if winsize % 2 == 0 else winsize

    # Reshape in case of nz=1
    im_arr = im_arr.reshape([nz, -1])
    if mask is not None:
        mask = mask.reshape([nz, -1])

    res_arr = []
    for i, im in enumerate(im_arr):
        # im should be a 1D array

        # Interpolate over masked data and NaN's
        nans = np.isnan(im)
        im_mask = nans if mask is None else nans | mask[i].flatten()
        if im_mask.any():
            # Create a copy so as to not change the original data
            im = np.copy(im)

            # Use a savgol filter to smooth out any outliers
            res = im.copy()
            res[~im_mask] = savgol_filter(im[~im_mask], 31, 3, mode='interp')

            # Replace masked pixels with linear interpolation
            x = mask_helper() # Returns the nonzero (True) indices of a mask
            im[im_mask]= np.interp(x(im_mask), x(~im_mask), res[~im_mask])

        if per_line:
            im = im.reshape([ny,-1])

            res = savgol_filter(im, winsize, order, axis=1, delta=1, **kwargs)
            res_arr.append(res)
        else:
            res = savgol_filter(im, winsize, order, delta=1, **kwargs)
            res_arr.append(res.reshape([ny,-1]))

    return np.array(res_arr).squeeze()


def channel_smooth_butter(im_arr, order=3, freq=0.1, per_line=False, mask=None):
    """Channel smoothing using Butterworth filter

    Parameters
    ==========
    im_arr : ndarray
        Input array of images (intended to be a cube of output channels). 
        Each image is operated on separately. If only two dimensions,
        then only a single input image is assumed.

    Keyword Args
    ============
    order : int
        Order of the filter (high order have sharper frequency cut-off)
    freq : float
        Normalized frequency cut-off (between 0 and 1). 1 is Nyquist.
    per_line : bool
        Smooth each channel line separately with the hopes of avoiding
        edge discontinuities.
    mask : bool image or None
        An image mask of pixels to ignore. Should be same size as im_arr.
        This can be used to mask pixels that the filter should ignore, 
        such as stellar sources or pixel outliers.
    """

    from scipy.signal import butter, filtfilt

    sh = im_arr.shape
    if len(sh)==2:
        nz = 1
        ny, chsize = sh
    else:
        nz, ny, chsize = sh

    # Reshape in case of nz=1
    im_arr = im_arr.reshape([nz, -1])
    if mask is not None:
        mask = mask.reshape([nz, -1])


    res_arr = []
    b, a = butter(order, freq, btype='lowpass', analog=False)
    for i, im in enumerate(im_arr):
        # im should be a 1D array

        # Interpolate over masked data and NaN's
        # Replace masked pixels with linear interpolation
        nans = np.isnan(im)
        im_mask = nans if mask is None else nans | mask[i].flatten()
        if im_mask.any():
            # Create a copy so as to not change the original data
            im = np.copy(im)

            # Use a savgol filter to smooth out any outliers
            res = im.copy()
            res[~im_mask] = savgol_filter(im[~im_mask], 31, 3, mode='interp')

            # Replace masked pixels with linear interpolation
            x = mask_helper() # Returns the nonzero (True) indices of a mask
            im[im_mask]= np.interp(x(im_mask), x(~im_mask), res[~im_mask])

        # Do filter line-by-line
        if per_line:
            im = im.reshape([ny,-1])

            res_lines = []
            for line in im:
                res = filtfilt(b, a, line)
                res_lines.append(res)
            res_lines = np.array(res_lines)
            res_arr.append(res_lines.reshape([ny,-1]))
        else:
            res = filtfilt(b, a, im)
            res_arr.append(res.reshape([ny,-1]))

    return np.array(res_arr).squeeze()






#from __future__ import absolute_import, division, print_function, unicode_literals

# import numpy as np
# from numpy.polynomial import legendre

from webbpsf_ext.maths import jl_poly, jl_poly_fit, fit_bootstrap
